import os

os.system('mkdir ./PANDORA_files')
os.system('mkdir ./PANDORA_files/data')
os.system('mkdir ./PANDORA_files/data/csv_pkl_files')
os.system('mkdir ./PANDORA_files/data/PDBs')
os.system('mkdir ./PANDORA_files/data/PDBs/pMHCI')
os.system('mkdir ./PANDORA_files/data/PDBs/pMHCII')
os.system('mkdir ./PANDORA_files/data/PDBs/Bad')
os.system('mkdir ./PANDORA_files/data/PDBs/Bad/pMHCI')
os.system('mkdir ./PANDORA_files/data/PDBs/Bad/pMHCII')
os.system('mkdir ./PANDORA_files/data/PDBs/IMGT_retrieved/')
os.system('mkdir ./test/test_data/PDBs/Bad')
os.system('mkdir ./test/test_data/PDBs/Bad/pMHCI')
os.system('mkdir ./test/test_data/PDBs/Bad/pMHCII')
os.system('mkdir ./test/test_data/csv_pkl_files')



# netMHCII Pan install

# wd = os.path.dirname(os.path.abspath(__file__))
# # Changing working directory
# os.chdir(wd + '/netMHCIIpan-4.0')
# # Downloading data folder for netMHCIIpan
# os.system('wget https://services.healthtech.dtu.dk/services/NetMHCIIpan-4.0/data.tar.gz')
# # Uncompressing
# os.system('gunzip data.tar.gz')
# os.system('tar -xvf data.tar')
# # remove tar file
# os.system('rm data.tar')
# # Make sure the tmp directory is created
# if not os.path.exists('tmp'):
#     os.mkdir('tmp')
#
# # Edit paths in the netMHCIIpan script
# netMHCIIpan_script = []
# with open('netMHCIIpan') as f:
#     for line in f:
#         netMHCIIpan_script.append(line)
# with open('netMHCIIpan', 'w') as f:
#     for line in netMHCIIpan_script:
#         if 'setenv\tNMHOME' in line:
#             f.write('setenv\tNMHOME\t' + wd + '/netMHCIIpan-4.0\n')
#             f.write('setenv\tTMPDIR\t' + wd + '/netMHCIIpan-4.0/tmp\n')
#         else:
#             f.write(line)
#
# # netMHC Pan install
#
# os.chdir(wd + '/netMHCpan-4.1')
# # Downloading data folder for netMHCIIpan
# os.system('wget https://services.healthtech.dtu.dk/services/NetMHCpan-4.1/data.tar.gz ')
# # Uncompressing
# os.system('gunzip data.tar.gz')
# os.system('tar -xvf data.tar')
# # remove tar file
# os.system('rm data.tar')
# # Make sure the tmp directory is created
# if not os.path.exists('tmp'):
#     os.mkdir('tmp')
#
# # Edit paths in the netMHCIIpan script
# netMHCpan_script = []
# with open('netMHCpan') as f:
#     for line in f:
#         netMHCpan_script.append(line)
# with open('netMHCpan', 'w') as f:
#     for line in netMHCpan_script:
#         if 'setenv\tNMHOME' in line:
#             f.write('setenv\tNMHOME\t' + wd + '/netMHCpan-4.1\n')
#             f.write('setenv\tTMPDIR\t' + wd + '/netMHCpan-4.1/tmp\n')
#         else:
#             f.write(line)
#
# # Changing working directory back
# os.chdir(wd)



from Bio import SeqIO
from Bio.Align.Applications import MuscleCommandline
from Bio.Align import substitution_matrices
from Bio.Align import PairwiseAligner
import Bio.Align
import PANDORA
import os


class Align:

    def __init__(self, target, template, 
                 output_dir=PANDORA.PANDORA_data + '/outputs', remove_terms=True):
        ''' Performs a alignment of the target and template(s). Will spit out a filename that will be used for modeller.

        Args:
            target: (Target object) The target object that will be aligned to 
                the template structures
            template: (Template object) can be either a single Target object 
                or a list of Target objects
            output_dir: (string) output directory
            remove_terms (bool): if True, removes N-terminal and C-terminal 
                regions not present in the template sequence
        '''

        # Assign target and template. If the template is not given as a list, put in in a list.
        self.__target = target
        if isinstance(template, list):
            self.__template = template
        else:
            self.__template = [template]
        self.__output_dir = output_dir # + '/%s_%s' %('_'.join([i.id for i in self.__template]), self.__target.id)

        # Find out if the target is MHC class I or II
        self.__MHC_class = target.MHC_class

        # Store the target and template ids for later use
        self.__tar_id = target.id
        self.__tem_id = [i.id for i in self.__template]

        # Define template m, n and p seqs
        self.__tem_m = [i.M_chain_seq for i in self.__template]
        if self.__MHC_class == 'II':
            self.__tem_n = [i.N_chain_seq for i in self.__template]
        self.__tem_p = [i.peptide for i in self.__template]

        # Define target m, n and p seqs. If there are no m and n chains supplied, just take the seqs from the template
        # Only if the target is MHC class II, the n chains are defined.
        if self.__target.M_chain_seq == '':
            self.__tar_m = self.__tem_m[0]
            if self.__MHC_class == 'II' and self.__target.N_chain_seq == '':
                self.__tar_n = self.__tem_n[0]
        else:
            self.__tar_m = self.__target.M_chain_seq
            if self.__MHC_class == 'II':
                self.__tar_n = self.__target.N_chain_seq
        self.__tar_p = self.__target.peptide

        # Perform alignment
        self.aligned_seqs_and_pept = self.__align_chains()
        # Cut extra N-terminal and C-terminal
        if remove_terms==True:
            self.__remove_terms()
        #Write alignment file for MODELLER
        self.alignment_file = self.__write_ali_file()


    def __align_chains(self):
        ''' Alignes the chains of MHCI/MHCII structues given a target and (multiple) templates.

        Returns: (dict) all aligned chains per structure id {id M: NNNNNNNNNNNN, id N: NNNNNNNNNNNN, id P: NNNNNN}

        '''

        # Align M and N chain for MHC II. Because the target chains need to be aligned to the respective chain of
        # the template, M and N are done seperately and later added together
        if self.__MHC_class == 'II':

            # Align the M chain
            # First write a fasta file containing all chains
            with open('%s/%s_M.fasta' % (self.__output_dir,self.__tar_id),"w") as f:
                for i in range(len(self.__tem_id)):
                    f.write('>'+self.__tem_id[i] + ' M\n' + self.__tem_m[i] +'\n')
                f.write('>' + self.__tar_id + ' M\n' + self.__tar_m)
            # Perform MSA with muscle
            msl = MuscleCommandline(input='%s/%s_M.fasta' % (self.__output_dir,self.__tar_id), out='%s/%s_M.afa' % (self.__output_dir, self.__tar_id))
            os.system(str(msl) + ' -quiet')

            # Align the N chain
            # First write a fasta file containing all chains
            with open('%s/%s_N.fasta' % (self.__output_dir,self.__tar_id),"w") as f:
                for i in range(len(self.__tem_id)):
                    f.write('>'+self.__tem_id[i] + ' N\n' + self.__tem_n[i] +'\n')
                f.write('>' + self.__tar_id + ' N\n' + self.__tar_n)
            # Perform MSA with muscle
            msl = MuscleCommandline(input='%s/%s_N.fasta' % (self.__output_dir, self.__tar_id), out='%s/%s_N.afa' % (self.__output_dir, self.__tar_id))
            os.system(str(msl) + ' -quiet')


            # Merge M and N chain into one file
            os.system('cat %s/%s_M.afa %s/%s_N.afa > %s/alignment.afa' % (
            self.__output_dir.replace(' ', '\\ '), self.__tar_id, self.__output_dir.replace(' ', '\\ '), self.__tar_id,
            self.__output_dir.replace(' ', '\\ ')))


        if self.__MHC_class == 'I':
            # Align the M chain
            # First write a fasta file containing all chains
            with open('%s/%s_M.fasta' % (self.__output_dir,self.__tar_id),"w") as f:
                for i in range(len(self.__tem_id)):
                    f.write('>'+self.__tem_id[i] + ' M\n' + self.__tem_m[i] +'\n')
                f.write('>' + self.__tar_id + ' M\n' + self.__tar_m)
            # Perform MSA with muscle
            msl = MuscleCommandline(input='%s/%s_M.fasta' % (self.__output_dir,self.__tar_id), out='%s/alignment.afa' % (self.__output_dir))
            os.system(str(msl) + ' -quiet')

        # Load aligned seqs into dict
        seqs = {v.description: str(v.seq) for (v) in SeqIO.parse('%s/alignment.afa' % (self.__output_dir), "fasta")}

        # Align peptides
        # aligned_pepts = {'1D9K P': 'GNSHRGAIEWEGIESG', '1IAK P': '-STDYGILQINSRW--'}
        aligned_pepts = self.__align_peptides()

        # Remove all intermediate files
        os.system('rm %s/*.fasta %s/*.afa' % (self.__output_dir.replace(' ', '\\ '), self.__output_dir.replace(' ', '\\ ')))
        return {**seqs, **aligned_pepts}

    def __align_peptides(self):
        ''' Align MHCI peptides based on their anchors. This function adds padding to the left and the right of
            the anchors and uses muscle to align the core. For MHCII, the peptides are aligned on the 2nd and 3rd anchor
            position. These two anchor positions are present in all MHCII structures (1 and 4 are missing sometimes).
            For MHCII, no Muscle alignment is performed because there is very little distance between the 2nd and 3rd
            anchor.

        Returns: (dict) {'id P': 'aligned_peptide_seq'}

        '''

        # Make a dict containing {id, (peptide_sequence, [anchors])}
        id_pept_anch = {self.__tar_id: (self.__tar_p, self.__target.anchors)}
        for i in self.__template:
            id_pept_anch[i.id] = (i.peptide, i.anchors)

        if self.__MHC_class == 'I':

            # Find the peptide with the longest 1st anchor.
            longest_1st_anch = max((v[1][0], k) for k, v in id_pept_anch.items())[0]
            # add padding to the left of all shorter peptides, so the 1st anchor of all peptide aligns
            for k, v in id_pept_anch.items():
                padding_len = longest_1st_anch - v[1][0]
                id_pept_anch[k] = ('-' * padding_len + v[0], [i + padding_len for i in v[1]])

            # align peptide cores
            cores = {k: v[0][v[1][0]:(v[1][1]) - 1] for k, v in id_pept_anch.items()}

            # Check if all cores are of equal length, otherwise align them
            if len(set([len(v) for k,v in cores.items()])) != 1:

                # Write Fasta file
                with open('%s/pept_cores.fasta' % (self.__output_dir), "w") as f:
                    for k, v in cores.items():
                        f.write('>' + k + '\n' + v + '\n')

                # Run Muscle
                msl = MuscleCommandline(input='%s/pept_cores.fasta' % (self.__output_dir),
                                        out='%s/pept_cores.afa' % (self.__output_dir))
                os.system(str(msl) + ' -quiet')

                # Load aligned seqs into dict
                aligned_cores = {v.description: str(v.seq) for (v) in
                                 SeqIO.parse('%s/pept_cores.afa' % (self.__output_dir), "fasta")}

                # Remove intermediate files
                # os.system(
                #     'rm %s/*.fasta %s/*.afa' % (
                #     self.__output_dir.replace(' ', '\\ '), self.__output_dir.replace(' ', '\\ ')))

                # Add aligned cores in between anchors
                for k, v in id_pept_anch.items():
                    id_pept_anch[k] = (v[0][:v[1][0]] + aligned_cores[k] + v[0][(v[1][1]) - 1:], v[1])


        if self.__MHC_class == 'II':

            # Find the peptide with the longest 2nd anchor.
            longest_2nd_anch = max((v[1][1], k) for k, v in id_pept_anch.items())[0]
            # add padding to the left of all shorter peptides, so the 2nd anchor of all peptide aligns
            for k, v in id_pept_anch.items():
                padding_len = longest_2nd_anch - v[1][1]
                id_pept_anch[k] = ('-' * padding_len + v[0], [i + padding_len for i in v[1]])

            # Find the longest 3rd anchor
            # add padding between the 2nd and 3rd anchor if needed to align these
            longest_3nd_anch = max((v[1][2], k) for k, v in id_pept_anch.items())[0]
            for k, v in id_pept_anch.items():
                padding_len = longest_3nd_anch - v[1][2]
                id_pept_anch[k] = (v[0][:v[1][2] - 1] + '-' * padding_len + v[0][v[1][2] - 1:],
                                   v[1][:2] + [i + padding_len for i in v[1][2:]])


        # Add padding to the right of the shortest peptides to match their length
        longest_peptide = max((len(v[0]), k) for k, v in id_pept_anch.items())[0]
        for k, v in id_pept_anch.items():
            padding_len = longest_peptide - len(v[0])
            id_pept_anch[k] = (v[0] + '-' * padding_len, v[1])

        # Reformat
        return {k + ' P': v[0] for k, v in id_pept_anch.items()}


    def __remove_terms(self):
        '''Removes N-terminal and C-terminal regions not present in the template'''
        if self.__MHC_class == 'I':
            chains = ['M']
        elif self.__MHC_class == 'II':
            chains = ['M', 'N']
            
        for chain in chains:
            templ_seq = self.aligned_seqs_and_pept['%s %s' %(self.__template[0].id, chain)]
            
            #N terminal excess
            N_term_excess = 0
            if templ_seq.startswith('-'):
                for aa in templ_seq:
                    if aa == '-':
                        N_term_excess += 1
                    else:
                        break

                #Cut N-terminal excess regions
                self.aligned_seqs_and_pept['%s %s' %(self.__template[0].id, chain)] = self.aligned_seqs_and_pept['%s %s' %(self.__template[0].id, chain)][N_term_excess:]
                self.aligned_seqs_and_pept['%s %s' %(self.__target.id, chain)] = self.aligned_seqs_and_pept['%s %s' %(self.__target.id, chain)][N_term_excess:]
                
                
            
            #C terminal excess
            C_term_excess = 0
            if templ_seq.endswith('-'):
                for aa in templ_seq[::-1]:
                    if aa == '-':
                        C_term_excess += 1
                    else:
                        break
            
                #Cut C-terminal excess regions
                self.aligned_seqs_and_pept['%s %s' %(self.__template[0].id, chain)] = self.aligned_seqs_and_pept['%s %s' %(self.__template[0].id, chain)][:-C_term_excess]
                self.aligned_seqs_and_pept['%s %s' %(self.__target.id, chain)] = self.aligned_seqs_and_pept['%s %s' %(self.__target.id, chain)][:-C_term_excess]
            print('Chain %s, excess N %i, C %i' %(chain, N_term_excess, C_term_excess))        
        

    def __write_ali_file(self):
        ''' Writes the alignement file from self.aligned_seqs_and_pept '''

        seqs = self.aligned_seqs_and_pept

        aligned_seqs = {} #dict for keeping all info that will be written to the .ali file
        for id in set([i.rsplit(' ', 1)[0] for i in seqs.keys()]): #Go through all structures

            # The header
            head = '>P1;'+id

            # The comment line that contains the path to the file, start chain and peptide length for modeller.
            if id in self.__tem_id:
                comment = 'structure:%s:1:M:%s:P::::' % (
                    os.path.basename(self.__template[self.__tem_id.index(id)].pdb_path), len(seqs[id + ' P']))
                    # self.__template[self.__tem_id.index(id)].pdb_path.replace(' ', '\\ '), len(seqs[id + ' P']))
            else:
                comment = 'sequence:::::::::'

            # The sequences
            if self.__MHC_class == 'II':
                seq = '%s/%s/%s*' %(seqs[id+' M'], seqs[id+' N'], seqs[id+' P'])
            if self.__MHC_class == 'I':
                seq = '%s/%s*' % (seqs[id + ' M'], seqs[id + ' P'])

            aligned_seqs[id] = (head, comment, seq)

        # Write actual .ali file
        alignment_file = '%s/%s.ali' %(self.__output_dir, self.__target.id)
        with open(alignment_file, 'w') as f:
            for k,v in aligned_seqs.items():
                f.write(v[0]+'\n'+v[1]+'\n'+v[2]+'\n\n')

        return alignment_file


class Align2:

    def __init__(self, target, template, output_dir=PANDORA.PANDORA_data + '/outputs'):
        ''' Performs a alignment of the target and template(s). Will spit out a filename that will be used for modeller.

        Args:
            target: (Target object) The target object that will be aligned to the template structures
            template: (Template object) can be either a single Target object or a list of Target objects
            output_dir: (string)
        '''

        self.target = target
        if isinstance(template, list):
            self.templates = template
        else:
            self.templates = [template]
        self.output_dir = output_dir



    def align_templates(self):
        ''' Alignes the chains of MHCI/MHCII structues given a target and (multiple) templates.

        Returns: (dict) all aligned chains per structure id {id M: NNNNNNNNNNNN, id N: NNNNNNNNNNNN, id P: NNNNNN}

        '''
        # Define some variables
        MHC_class = self.target.MHC_class
        opd = self.output_dir
        tar_id = self.target.id
        tem_id = [i.id for i in self.templates]

        tem_m = [i.M_chain_seq for i in self.templates]
        if MHC_class == 'II':
            tem_n = [i.N_chain_seq for i in self.templates]
        tem_p = [i.peptide for i in self.templates]

        tar_p = self.target.peptide
        if self.target.M_chain_seq == '':
            tar_m = tem_m[0]
            if MHC_class == 'II' and self.target.N_chain_seq == '':
                tar_n = tem_n[0]
        else:
            tar_m = self.target.M_chain_seq
            if MHC_class == 'II':
                tar_n = self.target.N_chain_seq


        # Align M and N chain for MHC II. Because the target chains need to be aligned to the respective chain of
        # the template, M and N are done seperately and later added together
        if MHC_class == 'II':

            # Align the M chain
            # First write a fasta file containing all chains
            with open('%s/%s_M.fasta' % (opd,tar_id),"w") as f:
                for i in range(len(tem_id)):
                    f.write('>'+tem_id[i] + ' M\n' + tem_m[i] +'\n')
                f.write('>' + tar_id + ' M\n' + tar_m)
            # Perform MSA with muscle
            msl = MuscleCommandline(input='%s/%s_M.fasta' % (opd,tar_id), out='%s/%s_M.afa' % (opd, tar_id))
            os.system(str(msl) + ' -quiet')

            # Align the N chain
            # First write a fasta file containing all chains
            with open('%s/%s_N.fasta' % (opd,tar_id),"w") as f:
                for i in range(len(tem_id)):
                    f.write('>'+tem_id[i] + ' N\n' + tem_n[i] +'\n')
                f.write('>' + tar_id + ' N\n' + tar_n)
            # Perform MSA with muscle
            msl = MuscleCommandline(input='%s/%s_N.fasta' % (opd, tar_id), out='%s/%s_N.afa' % (opd, tar_id))
            os.system(str(msl) + ' -quiet')


            # Merge M and N chain into one file
            os.system('cat %s/%s_M.afa %s/%s_N.afa > %s/alignment.afa' % (opd, tar_id, opd, tar_id, opd))


        if MHC_class == 'I':
            # Align the M chain
            # First write a fasta file containing all chains
            with open('%s/%s_M.fasta' % (opd,tar_id),"w") as f:
                for i in range(len(tem_id)):
                    f.write('>'+tem_id[i] + ' M\n' + tem_m[i] +'\n')
                f.write('>' + tar_id + ' M\n' + tar_m)
            # Perform MSA with muscle
            msl = MuscleCommandline(input='%s/%s_M.fasta' % (opd,tar_id), out='%s/alignment.afa' % (opd))
            os.system(str(msl) + ' -quiet')

        # Load aligned seqs into dict
        seqs = {v.description: str(v.seq) for (v) in SeqIO.parse('%s/alignment.afa' % (opd), "fasta")}

        # Align peptides
        # aligned_pepts = {'1D9K P': 'GNSHRGAIEWEGIESG', '1IAK P': '-STDYGILQINSRW--'}
        self.align_peptides()

        self.aligned_seqs = {**seqs, **self.aligned_pepts}

        # Remove all intermediate files
        os.system('rm %s/*.fasta %s/*.afa' % (opd, opd))

        # Using the aligned sequences, write the alignment file.
        self.alignment_file = self.__write_ali_file()


    def align_peptides(self, subt_matrix='PAM30'):
        ''' This function alignes all template peptides with the target peptide based on the anchors (pairwise).
            For MHCI, the left (L), anchors + middle (A,M) and right (R) part of the peptide are aligned and scored:
             LL AMMMMMA RR. This way the anchors always line up. In the end the three parts and their scores are added
             together.
            For MHCII, the same thing is done. However, instead the part between the 2nd and 3rd anchor is taken as as
            middle part. All IMGT structures have proper 2nd and 3rd anchors, while their 1st and especially 4th anchor
            can sometimes be debatable.

        Args:
            subt_matrix: Which subtitution matrix to use? PAM30, BLOSUM80, etc

        Returns: self.aligned_pepts and self.pept_alignment_scores

        '''

        # keep track of alignment scores
        scores = {}
        MHC_class = self.target.MHC_class
        tar_pept = self.target.peptide
        tar_anchors = self.target.anchors
        # Perform a pairwise alignment of the target and all templates for the MHC M chain and peptide
        for i in self.templates + [self.target]:
            temp_anchors = i.anchors

            if MHC_class == 'I':
                # Take the left (of anchor 1) middle (between anchor 1 and 2) and right (of anchor 2) of target peptide
                tar_left = tar_pept[:tar_anchors[0] - 1]
                tar_right = tar_pept[tar_anchors[1]:]
                tar_middle = tar_pept[tar_anchors[0] - 1:tar_anchors[1]]
                # Take the left (of anchor 1) middle (between anchor 1 and 2) and right (of anchor 2) of template peptide
                temp_left = i.peptide[:temp_anchors[0] - 1]
                temp_right = i.peptide[temp_anchors[1]:]
                temp_middle = i.peptide[temp_anchors[0] - 1:temp_anchors[1]]

            if MHC_class == 'II':
                # Take the left (of anchor 2) middle (between anchor 2 and 3) and right (of anchor 3) of target peptide
                tar_left = tar_pept[:tar_anchors[1] - 1]
                tar_right = tar_pept[tar_anchors[2]:]
                tar_middle = tar_pept[tar_anchors[1] - 1:tar_anchors[2]]
                # Take the left (of anchor 2) middle (between anchor 2 and 3) and right (of anchor 3) of template peptide
                temp_left = i.peptide[:temp_anchors[1] - 1]
                temp_right = i.peptide[temp_anchors[2]:]
                temp_middle = i.peptide[temp_anchors[1] - 1:temp_anchors[2]]

            # Create aligner object. The high gap penalties make sure gaps are only created if really needed.
            # For example if the anchors of target and template do not match and a gap needs to be introduced.
            aligner = Bio.Align.PairwiseAligner()
            aligner.substitution_matrix = substitution_matrices.load(subt_matrix)  # PAM30 for pept??
            aligner.gap_score = -1000
            aligner.end_open_gap_score = -1000000
            aligner.internal_open_gap_score = -10000

            # Aligns the middle part of the target and template peptide
            aligned_middle = aligner.align(tar_middle, temp_middle)
            aligned_middle_score = aligned_middle.score
            aligned_middle_seq = format([a for a in aligned_middle][0]).split('\n')[2]

            # Align residues left of the first anchors (if there are any)
            aligned_left_score = 0
            if len(tar_left) > 0 and len(temp_left) > 0:
                # Create aligner object. Make sure no gaps are created in the right side of the sequence
                aligner = Bio.Align.PairwiseAligner()
                aligner.substitution_matrix = substitution_matrices.load(subt_matrix)
                aligner.gap_score = -1000
                aligner.right_open_gap_score = -1000000
                aligner.internal_open_gap_score = -10000
                # Align the sequences
                aligned_left = aligner.align(tar_left, temp_left)
                aligned_left_score = aligned_left.score
                aligned_left_seq = format([a for a in aligned_left][0]).split('\n')[2]
            elif len(temp_left) > 0:
                aligned_left_seq = temp_left
            else:
                aligned_left_seq = ''

            # Align residues right of the second anchors (if there are any)
            aligned_right_score = 0
            if len(tar_right) > 0 and len(temp_right) > 0:
                # Create aligner object. Make sure no gaps are created in the left side of the sequence
                aligner = Bio.Align.PairwiseAligner()
                aligner.substitution_matrix = substitution_matrices.load(subt_matrix)
                aligner.gap_score = -1000
                aligner.left_open_gap_score = -1000000
                aligner.internal_open_gap_score = -10000
                # Align the sequences
                aligned_right = aligner.align(tar_right, temp_right)
                aligned_right_score = aligned_right.score
                aligned_right_seq = format([a for a in aligned_right][0]).split('\n')[2]
            elif len(temp_left) > 0:
                aligned_right_seq = temp_right
            else:
                aligned_right_seq = ''

            # Calculate the total similarity score by summing the scores of the left, middle and right part.
            total_score = aligned_left_score + aligned_middle_score + aligned_right_score
            # Paste the left, middle and right part of the template peptide together again.
            aligned_seq = aligned_left_seq + aligned_middle_seq + aligned_right_seq

            # add results to dict
            scores[i.id] = [total_score, aligned_seq, i.anchors]

        # Add padding left and right to make sure all peptides are the same length
        if MHC_class == 'I':
            scores = self.__add_padding(scores, first_aligned_anch_idx=0)
        if MHC_class == "II":
            scores = self.__add_padding(scores, first_aligned_anch_idx=1)

        # Dict that is later used in align_templates()
        self.aligned_pepts = {k + ' P': v[1] for k, v in scores.items()}

        # Remove the score of the self alignment between target.peptide and target.peptide. Also remove gaps
        scores.pop(self.target.id)
        # Dict of aligned peptides with their scores
        self.pept_alignment_scores = {k:(v[0],v[1], v[2]) for k,v in scores.items()}


    def __add_padding(self, scores, first_aligned_anch_idx=1):
        ''' Internal function for align_peptides() to add --padding-- left and right of peptides. Adds -- according to
            anchors positions

        Args:
            scores: dict: {'1FYT': (-2050.0, '---PKYVKQNTLKLAT--', [3, 6, 8, 11])
            first_aligned_anch_idx: int, 0 for MHCI and 1 for MHCII

        Returns: scores dict with padded peptides

        '''

        # add padding to the left of all shorter peptides, so the 1st (MHCI) or 2nd (MHCII) anchor of all peptide aligns
        longest_2nd_anch = max((v[2][first_aligned_anch_idx], k) for k, v in scores.items())[0]
        for k, v in scores.items():
            padding_len = longest_2nd_anch - v[2][first_aligned_anch_idx]
            scores[k] = (v[0], '-' * padding_len + v[1], [i + padding_len for i in v[2]])

        # Add padding to the right of the shortest peptides to match their length
        longest_peptide = max((len(v[1]), k) for k, v in scores.items())[0]
        for k, v in scores.items():
            padding_len = longest_peptide - len(v[1])
            scores[k] = (v[0], v[1] + '-' * padding_len, v[2])

        return scores

    def __write_ali_file(self):
        ''' Writes the alignement file from self.aligned_seqs_and_pept '''

        seqs = self.aligned_seqs
        tem_id = [i.id for i in self.templates]

        aligned_seqs = {} #dict for keeping all info that will be written to the .ali file
        for id in set([i.rsplit(' ', 1)[0] for i in seqs.keys()]): #Go through all structures

            # The header
            head = '>P1;'+id

            # The comment line that contains the path to the file, start chain and peptide length for modeller.
            if id in tem_id:
                comment = 'structure:%s:1:M:%s:P::::' % (
                    os.path.basename(self.templates[tem_id.index(id)].pdb_path), len(seqs[id + ' P']))
                    # self.__template[self.__tem_id.index(id)].pdb_path.replace(' ', '\\ '), len(seqs[id + ' P']))
            else:
                comment = 'sequence:::::::::'

            # The sequences
            if self.target.MHC_class == 'II':
                seq = '%s/%s/%s*' %(seqs[id+' M'], seqs[id+' N'], seqs[id+' P'])
            if self.target.MHC_class == 'I':
                seq = '%s/%s*' % (seqs[id + ' M'], seqs[id + ' P'])

            aligned_seqs[id] = (head, comment, seq)

        # Write actual .ali file
        alignment_file = '%s/%s.ali' %(self.output_dir, self.target.id)
        with open(alignment_file, 'w') as f:
            for k,v in aligned_seqs.items():
                f.write(v[0]+'\n'+v[1]+'\n'+v[2]+'\n\n')

        return alignment_file



from Bio.PDB import PDBParser
from Bio.PDB.Polypeptide import three_to_one
from Bio import pairwise2
import os
from Bio.PDB import PDBIO
import PANDORA
import traceback
#import sys
#import numpy as np
# from copy import deepcopy



# target = y.target
# model_path = y.model_path
# output_dir = PANDORA.PANDORA_data
# pdb = deepcopy(y.pdb)
# reference_pdb = '/Users/derek/Dropbox/Master_Bioinformatics/Internship/PANDORA/PANDORA_files/data/PDBs/pMHCII/1HXY.pdb'



class Model:

    def __init__(self, target, model_path='', output_dir = PANDORA.PANDORA_data, pdb=False, molpdf=0, dope=0):
        '''__init__(self, target, model_path='', output_dir = PANDORA.PANDORA_data, pdb=False, molpdf=0, dope=0)
         Initiate model object

        Args:
            target: Target object
            output_dir: (string) output directory
            model_path: (string) path to hypothetical model
            pdb:  Bio.PDB object of the hypothetical model
            molpdf: (float) molpdf score
            dope:  (float) DOPE score
        '''


        self.target = target
        self.model_path = model_path
        self.molpdf = molpdf
        self.dope = dope
        self.output_dir = output_dir

        # Check if the user gave either the path to the model pdb or the pdb itself.
        if self.model_path == '' and not pdb:
            raise Exception('Provide the path to a model structure or a Bio.PDB object')
        # If there is a model path and no pdb, parse the pdb structure from that path.
        if not pdb:
            self.pdb = PDBParser(QUIET=True).get_structure(self.target.id, self.model_path)


    def calc_LRMSD(self, reference_pdb, atoms = ['C', 'CA', 'N', 'O']):
        ''' Calculate the L-RMSD between the decoy and reference structure (ground truth).
            This function requires the pdb2sql module for L-RMSD calculation.

        Args:
            reference_pdb: Bio.PDB object or path to pdb file

        Returns: (float) L-RMSD
        '''

        #from pdb2sql import pdb2sql, superpose, StructureSimilarity
        from pdb2sql import StructureSimilarity

        # load target pdb
        if isinstance(reference_pdb, str):  # if its a string, it should be the path of the pdb, then load pdb first
            ref = PDBParser(QUIET=True).get_structure(self.target.id, reference_pdb)
        else:
            ref = reference_pdb

        # Define file names as variables
        #decoy_path = '%s/%s_decoy.pdb' % (self.output_dir, self.target.id)
        #ref_path = '%s/%s_ref.pdb' % (self.output_dir, self.target.id)

        # Define zones to align
        #M_lzone = list(range(4,73))
        #N_lzone = list(range(10,80))

        # pdb2sql needs 1 big chain and 1 ligand chain with correct numbering, for MHCII, this means merging the chains.
        model_name = self.model_path.split('/')[-1].split('.')[1]
        decoy_path, ref_path = homogenize_pdbs(self.pdb, ref, self.output_dir, model_name)

        start_dir = os.getcwd()
        os.chdir(self.output_dir)
        # Produce lzone file for the l-rmsd calculation
        #lzone = get_Gdomain_lzone(ref_path, self.output_dir, self.target.MHC_class)
        #TODO: check if it's MHC I or II and adapt for chain M and N
        # Get decoy structure to superpose
        #decoy_db = pdb2sql(decoy_path)
        #decoy_lzone = np.asarray(decoy_db.get('x,y,z', resSeq=M_lzone))

        # Get ref structure to superpose
        #ref_db = pdb2sql(ref_path)
        #ref_lzone = np.asarray(ref_db.get('x,y,z', resSeq=M_lzone))

        # Align the G domains
        #superpose.superpose_selection()

        try:
            # Calculate l-rmsd between decoy and reference with pdb2sql
            sim = StructureSimilarity(decoy_path, ref_path)
            #self.lrmsd = sim.compute_lrmsd_fast(method='svd', name=atoms, lzone = lzone)
            self.lrmsd = sim.compute_lrmsd_pdb2sql(exportpath=None, method='svd', name = atoms)
        except:
            print('An error occurred while calculating the rmsd for target %s, model %s' %(self.target.id, self.model_path))
            traceback.print_exc()
            raise Exception('Please check your model and ref info for model %s' %self.model_path)

        # remove intermediate files
        #os.system('rm %s/%s_decoy.pdb %s/%s_ref.pdb' %(self.output_dir, self.target.id, self.output_dir, self.target.id))
        #os.chdir(os.path.dirname(PANDORA.PANDORA_path))
        os.chdir(start_dir)

    def calc_Core_LRMSD(self, reference_pdb, atoms = ['C', 'CA', 'N', 'O']):
        ''' Calculate the L-RMSD between the decoy and reference structure (ground truth)

        Args:
            reference_pdb: Bio.PDB object or path to pdb file

        Returns: (float) L-RMSD
        '''

        from pdb2sql import StructureSimilarity

        # load target pdb
        if isinstance(reference_pdb, str):  # if its a string, it should be the path of the pdb, then load pdb first
            ref = PDBParser(QUIET=True).get_structure('MHC', reference_pdb)
        else:
            ref = reference_pdb

        # Define file names as variables
        decoy_path = '%s/%s_decoy.pdb' % (self.output_dir, self.target.id)
        ref_path = '%s/%s_ref.pdb' % (self.output_dir, self.target.id)

        # pdb2sql needs 1 big chain and 1 ligand chain with correct numbering, for MHCII, this means merging the chains.
        homogenize_pdbs(self.pdb, ref, self.output_dir, self.target.id, anchors = self.target.anchors)

        start_dir = os.getcwd()
        os.chdir(self.output_dir)
        # Produce lzone file for the l-rmsd calculation
        #lzone = get_Gdomain_lzone('%s/%s_ref.pdb' %(self.output_dir, self.target.id), self.output_dir, self.target.MHC_class)
        # Get decoy structure to superpose
        #decoy_db = psb2sql()

        # Calculate l-rmsd between decoy and reference with pdb2sql
        sim = StructureSimilarity(decoy_path, ref_path)
        self.core_lrmsd = sim.compute_lrmsd_pdb2sql(exportpath=None, method='svd', name=atoms)


        # remove intermediate files
        os.system('rm %s %s' %(decoy_path, ref_path))
        #os.chdir(os.path.dirname(PANDORA.PANDORA_path))
        os.chdir(start_dir)

    def calc_flanking_LRMSD(self, reference_pdb, atoms=['C', 'CA', 'N', 'O']):
        ''' Calculate the L-RMSD between the decoy and reference structure (ground truth)

        Args:
            reference_pdb: Bio.PDB object or path to pdb file

        Returns: (float) L-RMSD
        '''

        from pdb2sql import StructureSimilarity

        # load target pdb
        if isinstance(reference_pdb, str):  # if its a string, it should be the path of the pdb, then load pdb first
            ref = PDBParser(QUIET=True).get_structure('MHC', reference_pdb)
        else:
            ref = reference_pdb

        # Define file names as variables
        decoy_path = '%s/%s_decoy.pdb' % (self.output_dir, self.target.id)
        ref_path = '%s/%s_ref.pdb' % (self.output_dir, self.target.id)
        #
        # decoy_path = '%s/%s_decoy.pdb' % (output_dir, target.id)
        # ref_path = '%s/%s_ref.pdb' % (output_dir, target.id)

        # pdb2sql needs 1 big chain and 1 ligand chain with correct numbering, for MHCII, this means merging the chains.
        homogenize_pdbs(self.pdb, ref, self.output_dir, self.target.id, anchors=self.target.anchors, flanking=True)

        # homogenize_pdbs(pdb, ref, output_dir, target.id, anchors=target.anchors, flanking=True)

        os.chdir(self.output_dir)
        # Produce lzone file for the l-rmsd calculation
        # lzone = get_Gdomain_lzone('%s/%s_ref.pdb' %(self.output_dir, self.target.id), self.output_dir, self.target.MHC_class)
        # Get decoy structure to superpose
        # decoy_db = psb2sql()

        # Calculate l-rmsd between decoy and reference with pdb2sql
        sim = StructureSimilarity(decoy_path, ref_path)
        self.flanking_lrmsd = sim.compute_lrmsd_pdb2sql(exportpath=None, method='svd', name=atoms)

        # remove intermediate files
        os.system('rm %s %s' % (decoy_path, ref_path))
        os.chdir(os.path.dirname(PANDORA.PANDORA_path))


def merge_chains(pdb):
    ''' Merges two chains of MHCII to one chain. pdb2sql can only calculate L-rmsd with one chain.

    Args:
        pdb: Bio.PDB object

    Returns: Bio.PDB object with its M and N chain merged as M chain

    '''
    # Merge chains
    if 'N' in [chain.id for chain in pdb.get_chains()]:

        for j in pdb[0]['N'].get_residues():
            j.id = (j.id[0], j.id[1], 'M')
            pdb[0]['M'].add(j)

        for i in pdb.get_chains():
            for model in pdb:
                for chain in model:
                    if chain.id in ['N']:
                        model.detach_child(chain.id)
    return pdb


def renumber(pdb_ref, pdb_decoy):
    ''' aligns two pdb's and renumber them accordingly.

    Args:
        pdb_ref:   Bio.PDB object
        pdb_decoy: Bio.PDB object
        

    Returns: Bio.PDB objects with renumbered residues

    '''
    ref_sequences = [[chain.id, ('').join([three_to_one(res.resname) for res in chain])]
                      for chain in pdb_ref.get_chains()]
    ref_sequences.sort()
    decoy_sequences = [[chain.id, ('').join([three_to_one(res.resname) for res in chain])]
                      for chain in pdb_decoy.get_chains()]
    decoy_sequences.sort()
    
    assert(len(ref_sequences) == len(decoy_sequences))
    
    for ind in range(len(ref_sequences)):
        pair = pairwise2.align.globalxx(ref_sequences[ind][1], decoy_sequences[ind][1])[0]
        ref_sequences[ind][1]   = pair.seqA
        decoy_sequences[ind][1] = pair.seqB
        
    ref_sequences = [[seq[0],[i+1 for i,res in enumerate(seq[1]) if res != '-']] for seq in ref_sequences]
    decoy_sequences = [[seq[0],[i+1 for i,res in enumerate(seq[1]) if res != '-']] for seq in decoy_sequences]
    
    def assign(pdb, pdb_sequences):
        ''' Renumbers the pdb using aligned sequences. 

        Args:
            pdb_ref:   Bio.PDB object
            pdb_decoy: Bio.PDB object
            

        Returns: Bio.PDB objects with renumbered residues

        '''
        
        
        for chain in pdb.get_chains():
            for seq in pdb_sequences:
                if chain.id == seq[0]:
                    for ind, res in enumerate(chain):
                        res.id = ('X', seq[1][ind], res.id[2])
        for chain in pdb.get_chains():
            for res in chain:
                res.id = (' ', res.id[1], ' ')
        return pdb

    pdb_ref = assign(pdb_ref, ref_sequences)
    pdb_decoy = assign(pdb_decoy, decoy_sequences)

    return pdb_ref, pdb_decoy

# decoy = PDBParser(QUIET=True).get_structure('MHC', y.model_path)
#
#
# for i in decoy[0]['P']:
#     print(i.id)
#
# decoy = deepcopy(y.pdb)
# target_id = target.id
# anchors=target.anchors
# flanking=True

def homogenize_pdbs(decoy, ref, output_dir, target_id = 'MHC', anchors =False, flanking=False):
    ''' Make sure that the decoy and reference structure have the same structure sequences.

    Args:
        decoy: Bio.PDB object of the decoy structure
        ref: Bio.PDB object of the reference structure
        output_dir: (string) directory that is used to write intermediate files

    Returns: (tuple) Bio.PDB objects with the same structure sequence

    '''

    # If you give the anchors, the core L-RMSD will be calculated.
    # The peptide residues before and after the first and last anchor residue will be discarded.
    if anchors and not flanking:
        for x in range(len(decoy[0]['P'])):
            for i in decoy[0]['P']:
                if i.id[1] < anchors[0] or i.id[1] > anchors[-1]:
                    decoy[0]['P'].detach_child(i.id)
            for i in ref[0]['P']:
                if i.id[1] < anchors[0] or i.id[1] > anchors[-1]:
                    ref[0]['P'].detach_child(i.id)

    # If you give the anchors AND flanking = True, the flanking L-RMSD will be calculated. Only if the peptide is
    # also longer than the binding core. The peptide binding core will be discarded
    if anchors and flanking and len(decoy[0]['P']) > 9:
        for x in range(len(decoy[0]['P'])):
            for i in decoy[0]['P']:
                if i.id[1] >= anchors[0] and i.id[1] <= anchors[-1]:
                    decoy[0]['P'].detach_child(i.id)
            for i in ref[0]['P']:
                if i.id[1] >= anchors[0] and i.id[1] <= anchors[-1]:
                    ref[0]['P'].detach_child(i.id)



    # remove c-like domain and keep only g domain
    decoy = remove_C_like_domain(decoy)
    ref = remove_C_like_domain(ref)

    # merge chains of the decoy
    decoy = merge_chains(decoy)
    # merge chains of the reference
    ref = merge_chains(ref)
    ref, decoy = renumber(ref, decoy)

    # Write pdbs
    decoy_path = '%s/%s_decoy.pdb' % (output_dir, target_id)
    io = PDBIO()
    io.set_structure(decoy)
    io.save('%s/%s_decoy.pdb' % (output_dir, target_id))

    ref_path = '%s/%s_ref.pdb' % (output_dir, target_id)
    io = PDBIO()
    io.set_structure(ref)
    io.save('%s/%s_ref.pdb' % (output_dir, target_id))

    return decoy_path, ref_path

def get_Gdomain_lzone(ref_pdb, output_dir, MHC_class):
    """ Produce a lzone file for pdb2sql.

    Args:
        ref_pdb (str): path to the pdb file to use for the lzone
        output_dir (str): output directory
        MHC_class (str): Class of the MHC

    Raises:
        Exception: In case there are unexpected chain names it raises an exception

    Returns:
        outfile (str): Path to the output file
    """

    ref_name = ref_pdb.split('/')[-1].split('.')[0]
    outfile = '%s/%s.lzone' %(output_dir, ref_name)
    if MHC_class == 'I':
        with open(outfile, 'w') as output:
            P = PDBParser(QUIET=1)
            structure = P.get_structure('r', ref_pdb)
            for chain in structure.get_chains():
                if chain.id == 'M':
                    for x in range(2,173):
                        output.write('zone %s%i-%s%i\n' %(chain.id, x, chain.id, x))
                    #output.write('zone %s2-%s172\n' %(chain.id, chain.id))
                    #output.write('zone %s2-%s172:%s2-%s172\n' %(chain.id, chain.id, chain.id, chain.id))
                elif chain.id == 'P':
                    pass
                    #output.write('fit\n')
                    #for residue in chain:
                    #    if residue.id[2] == ' ':
                    #        output.write('rzone %s%s-%s%s\n' %(chain.id, str(residue.id[1]), chain.id, str(residue.id[1])))
                else:
                    raise Exception('Unrecognized chain ID, different from M or P. Please check your file')
            #output.write('fit\n')

    elif MHC_class == 'II':
        #Chain M from 4 to 72; Chain N from 10 to 80
        with open(outfile, 'w') as output:
            P = PDBParser(QUIET=1)
            structure = P.get_structure('r', ref_pdb)
            for chain in structure.get_chains():
                if chain.id == 'M':
                    output.write('zone %s4-%s72:%s4-%s72\n' %(chain.id, chain.id, chain.id, chain.id))
                elif chain.id == 'N':
                    output.write('zone %s10-%s80:%s10-%s80\n' %(chain.id, chain.id, chain.id, chain.id))
                elif chain.id == 'P':
                    pass
                    #output.write('fit\n')
                    #for residue in chain:
                    #    if residue.id[2] == ' ':
                    #        output.write('rzone %s%s-%s%s\n' %(chain.id, str(residue.id[1]), chain.id, str(residue.id[1])))
                else:
                    raise Exception('Unrecognized chain ID, different from M, N or P. Please check your file')
            #output.write('fit\n')
    return outfile

def remove_C_like_domain(pdb):
    '''Removes the C-like domain from a MHC struture and keeps only the G domain
    Args:
        pdb: (Bio.PDB): Bio.PDB object with chains names M (N for MHCII) and P
    Returns: (Bio.PDB): Bio.PDB object without the C-like domain
    '''

    # If MHCII, remove the C-like domain from the M-chain (res 80 and higher) and the N-chain (res 90 and higher)
    if 'N' in [chain.id for chain in pdb.get_chains()]:

        residue_ids_to_remove_N = [res.id for res in pdb[0]['N'] if res.id[1] > 90]
        #  Remove them
        for id in residue_ids_to_remove_N:
            pdb[0]['N'].detach_child(id)

        residue_ids_to_remove_M = [res.id for res in pdb[0]['M'] if res.id[1] > 80]
        #  Remove them
        for id in residue_ids_to_remove_M:
            pdb[0]['M'].detach_child(id)

    # If MHCI, remove the C-like domain, which is from residue 180+
    if 'N' not in [chain.id for chain in pdb.get_chains()]:
        for chain in pdb.get_chains():
            if chain.id == 'M':
                need_to_be_removed = [res.id for res in chain if res.id[1] > 180]
                _ = [chain.detach_child(x) for x in need_to_be_removed]
    return pdb

#ValueError: Invalid column name lzone. Possible names are
#['rowID', 'serial', 'name', 'altLoc', 'resName', 'chainID', 'resSeq',
# 'iCode', 'x', 'y', 'z', 'occ', 'temp', 'element', 'model']



import pytest
import os
import PANDORA
from PANDORA.Contacts import Contacts
from PANDORA.Pandora import Align
from PANDORA.Database import Database
from PANDORA.Database import Database_functions
from PANDORA.PMHC import PMHC
from PANDORA.Pandora import Pandora
from PANDORA.PMHC import Model
#from PANDORA.Wrapper import Wrapper

def test_PMHC_target():
    # Create target object
    target = PMHC.Target('1A1O',
                         ['HLA-B*5301', 'HLA-B*5301'],
                         'KPIVQYDNF',
                         M_chain_seq = 'GSHSMRYFYTAMSRPGRGEPRFIAVGYVDDTQFVRFDSDAASPRTEPRPPWIEQEGPEYWDRNTQIFKTNTQTYRE'
                                       'NLRIALRYYNQSEAGSHIIQRMYGCDLGPDGRLLRGHDQSAYDGKDYIALNEDLSSWTAADTAAQITQRKWEAARV'
                                       'AEQLRAYLEGLCVEWLRRYLENGKETLQRADPPKTHVTHHPVSDHEATLRCWALGFYPAEITLTWQRDGEDQTQDT'
                                       'ELVETRPAGDRTFQKWAAVVVPSGEEQRYTCHVQHEGLPKPLTLRWEP',
                         anchors = [2, 9])
    # test if the inital model is empty, MHC class, the peptide chain, the allele and anchors
    pass_test = False
    if target.initial_model == False and target.MHC_class == 'I' and target.peptide == 'KPIVQYDNF':
        if target.peptide == 'KPIVQYDNF' and target.anchors == [2, 9] :
            pass_test = True

    assert pass_test


def test_PMHC_template():
    # Create template object
    template = PMHC.Template('1A1O',
                         ['HLA-B*5301', 'HLA-B*5301'],
                         pdb_path=PANDORA.PANDORA_path + '/../test/test_data/PDBs/pMHCI/1A1O.pdb')
    # check calculated anchors, M_chain, peptide, allele and pdb chains
    pass_test = False
    if template.anchors == [2,9] and template.M_chain_seq != '' and template.peptide == 'KPIVQYDNF':
        if template.allele_type == ['HLA-B*5301', 'HLA-B*5301']:
            if [i.id for i in template.pdb.get_chains()] == ['M', 'P']:
                pass_test = True

    assert pass_test


def test_fail_PMHC():
    # Try to fail initiating a template object
    pass_test = False
    try:
        PMHC.Template('id', ['allele'])
    except:
        pass_test = True

    assert pass_test


def test_contacts():
    # Calculate atom contacts
    c = Contacts.Contacts(PANDORA.PANDORA_path + '/../test/test_data/PDBs/pMHCI/1A1O.pdb')
    # check length of output, chains of the pdb and the pdb path
    pass_test = False
    if len(c.chain_contacts[0]) == 9 and [i.id for i in c.PDB.get_chains()] == ['M', 'P'] and c.pdb_path != '':
        pass_test = True

    assert pass_test


@pytest.mark.skip
def test_align():
    # initiate target and template object
    template = PMHC.Template('1A1O',
                         ['HLA-B*5301', 'HLA-B*5301'],
                         pdb_path=PANDORA.PANDORA_path + '/../test/test_data/PDBs/pMHCI/1A1O.pdb',
                             anchors = [2, 9])
    target = PMHC.Target('2X4R',
                         ['test'],
                         'NLVPMVATV',
                         M_chain_seq = 'GSHSMRYFFTSVSRPGRGEPRFIAVGYVDDTQFVRFDSDAASQRMEPRAPWIEQEGPEYWDGETRKVKAHSQTHRV'
                                       'DLGTLRGYYNQSEAGSHTVQRMYGCDVGSDWRFLRGYHQYAYDGKDYIALKEDLRSWTAADMAAQTTKHKWEAAHV'
                                       'AEQLRAYLEGTCVEWLRRYLENGKETLQRTDAPKTHMTHHAVSDHEATLRCWALSFYPAEITLTWQRDGEDQTQDT'
                                       'ELVETRPAGDGTFQKWAAVVVPSGQEQRYTCHVQHEGLPKPLTLRWE',
                         anchors = [2, 9])
    # align target and template
    a = Align.Align(target, template, output_dir=PANDORA.PANDORA_path + '/../test/')
    # check keys of aligned output and output .ali file path
    pass_test = False
    if a.aligned_seqs_and_pept['1A1O P'] == 'KPIVQYDNF' and '2X4R M' in a.aligned_seqs_and_pept:
        if os.path.basename(a.alignment_file) == '2X4R.ali':
            pass_test = True
    # Remove output dir
    os.system('rm -r %s1A1O_2X4R' %(PANDORA.PANDORA_path + '/../test/'))

    assert pass_test


def test_clean_MHCI_structure():
    x = Database_functions.parse_pMHCI_pdb('1A1O',
                       indir=PANDORA.PANDORA_path + '/../test/test_data/PDBs/IMGT_retrieved/IMGT3DFlatFiles',
                       outdir=PANDORA.PANDORA_path + '/../test/test_data/PDBs/pMHCI',
                       bad_dir=PANDORA.PANDORA_path + '/../test/test_data/PDBs/pMHCI/Bad')

    assert x.peptide == 'KPIVQYDNF' and [i.id for i in x.pdb.get_chains()] == ['M', 'P']


def test_clean_MHCII_structure():
    x = Database_functions.parse_pMHCII_pdb('2NNA',
                       indir=PANDORA.PANDORA_path + '/../test/test_data/PDBs/IMGT_retrieved/IMGT3DFlatFiles',
                       outdir=PANDORA.PANDORA_path + '/../test/test_data/PDBs/pMHCII',
                       bad_dir=PANDORA.PANDORA_path + '/../test/test_data/PDBs/pMHCII/Bad')


    assert x.peptide == 'SGEGSFQPSQENP' and [i.id for i in x.pdb.get_chains()] == ['M', 'N', 'P']


@pytest.mark.skip
def test_construct_database():
    test_data = PANDORA.PANDORA_path + '/../test/test_data/'
    bad1, bad2 = test_data + 'PDBs/Bad/pMHCI/6C6A.pdb', test_data + 'PDBs/Bad/pMHCII/1K8I.pdb'
    log1, log2 = test_data + 'PDBs/Bad/log_MHCI.csv', test_data + 'PDBs/Bad/log_MHCII.csv'

    # Construct database object
    db = Database.Database()
    db.construct_database(save=False, download=False, data_dir=test_data)

    # test the creation of bad files, log files and the information inside of the database object
    pass_test = False
    if os.path.exists(bad1) and os.path.exists(log2):
        if '1A1O' in db.MHCI_data and '4Z7U' in db.MHCII_data:
            if db.MHCII_data['2NNA'].peptide == 'SGEGSFQPSQENP' and db.MHCI_data['2X4R'].peptide == 'NLVPMVATV':
                pass_test = True

    # remove bad files
    os.system('rm %s %s %s %s' %(bad1, bad2, log1, log2))

    assert pass_test


def test_load_db():
    db = Database.load(PANDORA.PANDORA_path + '/../test/test_data/Test_Pandora_MHCI_and_MHCII_data.pkl')
    # test if items in the database are correct
    pass_test = False
    if '1A1O' in db.MHCI_data and '4Z7U' in db.MHCII_data:
        if db.MHCII_data['2NNA'].peptide == 'SGEGSFQPSQENP' and db.MHCI_data['2X4R'].peptide == 'NLVPMVATV':
            pass_test = True

    assert pass_test


def test_template_select_MHCI():
    db = Database.load(PANDORA.PANDORA_path + '/../test/test_data/Test_Pandora_MHCI_and_MHCII_data.pkl')
    # Create target object
    target = PMHC.Target('1A1O',
                         db.MHCI_data['1A1O'].allele_type,
                         db.MHCI_data['1A1O'].peptide,
                         M_chain_seq=db.MHCI_data['1A1O'].M_chain_seq,
                         anchors=db.MHCI_data['1A1O'].anchors)

    # Perform modelling
    mod = Pandora.Pandora(target, db, output_dir = os.path.dirname(PANDORA.PANDORA_path) + '/test')
    mod.find_template(benchmark=True)

    assert mod.template.id == '2X4R' and mod.template.peptide == 'NLVPMVATV'


def test_template_select_MHCII():
    # Load database
    db = Database.load(PANDORA.PANDORA_path + '/../test/test_data/Test_Pandora_MHCI_and_MHCII_data.pkl')
    # Create target object
    target = PMHC.Target('2NNA',
                         db.MHCII_data['2NNA'].allele_type,
                         db.MHCII_data['2NNA'].peptide,
                         MHC_class= 'II',
                         M_chain_seq=db.MHCII_data['2NNA'].M_chain_seq,
                         N_chain_seq=db.MHCII_data['2NNA'].N_chain_seq,
                         anchors=db.MHCII_data['2NNA'].anchors)

    # Perform modelling
    mod = Pandora.Pandora(target, db, output_dir = os.path.dirname(PANDORA.PANDORA_path) + '/test')
    mod.find_template(benchmark=True)

    assert mod.template.id == '4Z7U' and mod.template.peptide == 'PSGEGSFQPSQENPQ'

@pytest.mark.skip
def test_database_repath():
    # Load database
    db = Database.load(PANDORA.PANDORA_path + '/../test/test_data/Test_Pandora_MHCI_and_MHCII_data.pkl')
    # Repath database
    db.repath(PANDORA.PANDORA_path + '/../test/test_data/PDBs', save=PANDORA.PANDORA_path + '/../test/test_data/Test_Pandora_MHCI_and_MHCII_data_repath.pkl')
    
    MHCI_flag = False
    if db.MHCI_data['1A1O'].pdb_path == PANDORA.PANDORA_path + '/../test/test_data/PDBs/pMHCI/1A1O.pdb':
        MHCI_flag = True
    MHCI_flag = False
    if db.MHCII_data['2NNA'].pdb_path == PANDORA.PANDORA_path + '/../test/test_data/PDBs/pMHCII/2NNA.pdb':
        MHCII_flag = True
    assert  MHCI_flag and MHCII_flag
    
@pytest.mark.skip
def test_pandora_MHCI_modelling():
    # Load database
    db = Database.load(PANDORA.PANDORA_path + '/../test/test_data/Test_Pandora_MHCI_and_MHCII_data_repath.pkl')
    # Create target object
    target = PMHC.Target('1A1O',
                         db.MHCI_data['1A1O'].allele_type,
                         db.MHCI_data['1A1O'].peptide,
                         M_chain_seq=db.MHCI_data['1A1O'].M_chain_seq,
                         anchors=db.MHCI_data['1A1O'].anchors)

    # Perform modelling
    mod = Pandora.Pandora(target, db, output_dir = os.path.dirname(PANDORA.PANDORA_path) + '/test/test_output/')
    mod.model(n_loop_models=1, stdev=0.1, benchmark=True, loop_refinement='very_fast')

    # Check if mod.template is initiated and if the initial model is created. Then checks molpdf of output.
    pass_test = False
    if mod.template.id == '2X4R' and [c.id for c in mod.target.initial_model.get_chains()] == ['M', 'P']:
        if float(mod.results[0].moldpf) < 2000 and float(mod.results[0].moldpf) > 0:
            pass_test = True
    # remove output file
    os.system('rm -r %s' % (mod.output_dir))

    assert pass_test

@pytest.mark.skip
def test_wrapper_MHCI():
    # Load database
    db = Database.load(PANDORA.PANDORA_path + '/../test/test_data/Test_Pandora_MHCI_and_MHCII_data_repath.pkl')
    # Create Wrapper object
    wrap = Wrapper.Wrapper()
    # Define data_file
    data_file =  PANDORA.PANDORA_path + '/../test/test_data/test_MHCI_datafile.tsv'
    # Create targets
    wrap.create_targets(data_file, db, MHC_class='I', header=True, 
                        delimiter='/t', IDs_col=0, peptides_col=1, 
                        allele_col=3, anchors_col=2)
    # Check if the jobs went as expected
    
    # Define output directory
    output_dir = os.path.dirname(PANDORA.PANDORA_path) + '/test'
    # Run the modellings
    wrap.run_pandora(num_cores=2, n_loop_models=20, 
                     benchmark=False, output_dir=output_dir)
    
@pytest.mark.skip
def test_pandora_MHCII_modelling():
    # Load database
    db = Database.load(PANDORA.PANDORA_path + '/../test/test_data/Test_Pandora_MHCI_and_MHCII_data_repath.pkl')
    # Create target object
    target = PMHC.Target('2NNA',
                         db.MHCII_data['2NNA'].allele_type,
                         db.MHCII_data['2NNA'].peptide,
                         MHC_class= 'II',
                         M_chain_seq=db.MHCII_data['2NNA'].M_chain_seq,
                         N_chain_seq=db.MHCII_data['2NNA'].N_chain_seq,
                         anchors=db.MHCII_data['2NNA'].anchors)

    # Perform modelling
    mod = Pandora.Pandora(target, db, output_dir = os.path.dirname(PANDORA.PANDORA_path) + '/test/test_output/')
    mod.model(n_models=1, stdev=0.2, benchmark=True, loop_refinement='very_fast')

    # Check if mod.template is initiated and if the initial model is created. Then checks molpdf of output.
    pass_test = False
    if mod.template.id == '4Z7U' and [c.id for c in mod.target.initial_model.get_chains()] == ['M','N', 'P']:
        if float(mod.results[0].moldpf) < 1000 and float(mod.results[0].moldpf) > -1000:
            pass_test = True
    # remove output file
    os.system('rm -r %s' % (mod.output_dir))
    assert pass_test

def test_rmsd():
    # Load database
    db = Database.load(PANDORA.PANDORA_path + '/../test/test_data/Test_Pandora_MHCI_and_MHCII_data.pkl')
    # Create target object
    target = PMHC.Target('1A1O',
                         db.MHCI_data['1A1O'].allele_type,
                         db.MHCI_data['1A1O'].peptide,
                         M_chain_seq=db.MHCI_data['1A1O'].M_chain_seq,
                         anchors=db.MHCI_data['1A1O'].anchors)

    # Initiate Model object
    m = Model.Model(target, model_path=PANDORA.PANDORA_path + '/../test/test_data/1A1O.BL00010001.pdb',
                    output_dir=PANDORA.PANDORA_path + '/../test')
    # Calculate L-RMSD and Core L-RMSD
    m.calc_LRMSD(PANDORA.PANDORA_path + '/../test/test_data/PDBs/pMHCI/1A1O.pdb')
    m.calc_Core_LRMSD(PANDORA.PANDORA_path + '/../test/test_data/PDBs/pMHCI/1A1O.pdb')
    # Check if the rmsds are between 0.5 and 2 (I gave some slack for the cases that modeller gets lucky.
    pass_test = m.lrmsd > 1 and m.lrmsd < 1.5 and m.core_lrmsd > 1 and m.core_lrmsd < 1.5

    assert pass_test




from Bio.PDB import PDBParser
from Bio.SeqUtils import seq1
import PANDORA
from PANDORA.Contacts import Contacts
from PANDORA.Database import Database_functions
from PANDORA.Pandora import Modelling_functions
from PANDORA.PMHC import Anchors
from abc import ABC, abstractmethod
import os

class PMHC(ABC):

    def __init__(self, id, allele_type, peptide = '', MHC_class = 'I', M_chain_seq = '', N_chain_seq = '', anchors = [],
                 helix=False, sheet=False):
        ''' pMHC class. Acts as a parent class to Template and Target

        Args:
            id: (string) PDB identifier
            allele_type: (list) list of MHC alleles (or allele)
            peptide: (string) peptide sequence
            MHC_class: (string) either 'I' or 'II' denoting MHC class I and MHC class II respectively
            M_chain_seq: (string) M chain sequence for the Alpha chain
            N_chain_seq: (string) N chain sequence for the Beta chain
            anchors:  (list) list of integers specifying which residue(s) of the peptide should be fixed as an anchor
                        during the modelling. MHC class I typically has 2 anchors, while MHC class II typically has 4.
        '''
        super().__init__()
        self.id = id
        self.MHC_class = MHC_class
        self.peptide = peptide
        self.M_chain_seq = M_chain_seq
        self.N_chain_seq = N_chain_seq
        self.allele_type = allele_type
        self.anchors = anchors
        self.helix = helix
        self.sheet = sheet


        @abstractmethod
        def info(self):
            pass

        @abstractmethod
        def calc_contacts(self):
            pass

        @abstractmethod
        def calc_anchor_contacts(self):
            pass


class Template(PMHC):

    def __init__(self, id, allele_type, peptide='', MHC_class='I', M_chain_seq='', N_chain_seq='', anchors=[], helix=False, sheet=False, pdb_path=False, pdb=False, resolution=None):
        ''' Template structure class. This class holds all information of a template structure that is used for
            homology modelling. This class needs a id, allele and the path to a pdb file to work. (sequence info of
            the chains and peptide can be fetched from the pdb)

        Args:
            id: (string) PDB identifier
            allele_type: (list) list of MHC alleles (or allele)
            peptide: (string) peptide sequence
            MHC_class:  (string) either 'I' or 'II' denoting MHC class I and MHC class II respectively
            M_chain_seq: (string) M chain sequence for the Alpha chain
            N_chain_seq: (string) N chain sequence for the Beta chain
            anchors: (list) list of integers specifying which residue(s) of the peptide should be fixed as an anchor
                        during the modelling. MHC class I typically has 2 anchors, while MHC class II typically has 4.
            pdb_path: (string) path to pdb file
            pdb: (Bio.PDB) Biopython PBD object
            resolution: (float) Structure resolution in Angstrom
        '''
        super().__init__(id, allele_type, peptide, MHC_class, M_chain_seq, N_chain_seq, anchors, helix, sheet)
        self.pdb_path = pdb_path
        self.pdb = pdb
        self.contacts = False
        self.resolution = resolution


        if not pdb_path and not pdb:
            raise Exception('Provide a PDB structure to the Template object first')

        if pdb_path and not pdb: # If the path to a pdb file or a Bio.PDB object is given, parse the pdb
            self.parse_pdb()

        if anchors == []:
            self.calc_anchors()

    def parse_pdb(self):
        '''Loads pdb from path, updates self.pdb field and self.chain_seq/self.peptide if they were empty'''

        if self.pdb_path and not self.pdb: #if there is a path to a pdb provided and there is not already a self.pdb...
            parser = PDBParser(QUIET=True)  # Create a parser object, used to read pdb files
            self.pdb = parser.get_structure('MHC', self.pdb_path) #Create Bio.PDB object
            self.resolution = Database_functions.get_resolution(self.pdb_path) #Get resolution from pdb file

        # If the chains or peptide are not given by the user, fetch them from the pdb
        # Get the chain sequences
        chain_seqs = [seq1(''.join([res.resname for res in chain])) for chain in self.pdb.get_chains()]
        # Update chain and peptide fields if emtpy
        if self.MHC_class == 'I':
            if self.M_chain_seq == '':
                self.M_chain_seq = chain_seqs[0]
            if not self.peptide:
                self.peptide = chain_seqs[-1]
        if self.MHC_class == 'II':
            if self.M_chain_seq == '' and self.N_chain_seq == '':
                self.M_chain_seq = chain_seqs[0]
                self.N_chain_seq = chain_seqs[1]
            if not self.peptide:
                self.peptide = chain_seqs[-1]

    def info(self):
        """ Print the basic info of this structure

        """
        print('This is a %s structure.' %(type(self).__name__))
        print('ID: %s' %self.id)
        print('Type: MHC class %s' %self.MHC_class)
        print('Alleles: %s' % self.allele_type)
        if self.M_chain_seq != '':
            print('Alpha chain length: %s' %len(self.M_chain_seq))
        if self.N_chain_seq != '' and self.MHC_class == 'II':
            print('Beta chain length: %s' %len(self.N_chain_seq))
        print('Peptide length: %s' %len(self.peptide))
        if self.M_chain_seq != '':
            print('Alpha chain: %s' % self.M_chain_seq)
        if self.N_chain_seq != '':
            print('Beta chain: %s' % self.N_chain_seq)

        print('Peptide: %s' % self.peptide)
        print('Anchors: %s' %self.anchors)

        if self.sheet:
            print('Beta-sheet: %s' % self.sheet)
        if self.helix:
            print('Alpha-helix: %s' % self.helix)
        if self.pdb_path:
            print('Path to PDB file: %s' %self.pdb_path)
        if not self.pdb:
            print('PDB structure: no PDB structure provided')
        else:
            print('PDB structure:')
            for (k, v) in self.pdb.header.items():
                print('\t'+k + ':', v)

    def calc_contacts(self):
        if self.pdb:
            self.contacts = Contacts.Contacts(self.pdb)
        else:
            raise Exception('Provide a PDB structure to the Template object first')

    def calc_anchors(self):
        if self.MHC_class == 'I':
            self.anchors = Anchors.pMHCI_anchors(self.pdb)
        if self.MHC_class == 'II':
            self.anchors = Anchors.pMHCII_anchors(self.pdb)

    def calc_anchor_contacts(self):
        if self.pdb and self.anchors:
            self.anchor_contacts = Contacts.Contacts(self.pdb, anchors=self.anchors).anchor_contacts
        else:
            raise Exception('Provide an initial model (.ini PDB) and anchor positions to the Target object first')




class Target(PMHC):

    def __init__(self, id, allele_type, peptide, MHC_class = 'I', M_chain_seq = '', N_chain_seq = '', anchors = [],
                 helix=False, sheet=False, templates = False, use_netmhcpan = False):
        ''' Target structure class. This class needs an ID (preferably a PDB ID), allele and pepide information.

        Args:
            id: (string) PDB identifier
            peptide: (string) peptide sequence
            allele_type: (list) list of MHC alleles (or allele)
            MHC_class: (string) either 'I' or 'II' denoting MHC class I and MHC class II respectively
            M_chain_seq: (string) M chain sequence for the Alpha chain
            N_chain_seq: (string) N chain sequence for the Beta chain
            anchors: (list) list of integers specifying which residue(s) of the peptide should be fixed as an anchor
                        during the modelling. MHC class I typically has 2 anchors, while MHC class II typically has 4.
            templates: Template object. The user can specify that PANDORA uses a certain structure as template.
            use_netmhcpan (bool): If True, uses local installation of NetMHCPan to predict the anchors when
                                  anchor positions are not provided. Defaults to False.
        '''

        super().__init__(id, allele_type, peptide, MHC_class, M_chain_seq, N_chain_seq, anchors, helix, sheet)
        self.templates = templates
        self.initial_model = False
        self.contacts = False
        self.anchor_contacts = False

        # If the user does provide sequence info, make sure both the M and N chain are provided
        if MHC_class == 'II' and M_chain_seq != '' and N_chain_seq == '':
            raise Exception('Provide both the M and N chain sequences for MHC class II targets or none at all')
        if MHC_class == 'II' and N_chain_seq != '' and M_chain_seq == '':
            raise Exception('Provide both the M and N chain sequences for MHC class II targets or none at all')

        # If the user does not provide sequence info, retrieve them from the reference sequences.
        # WARNING: currently available only for MHC I
        if MHC_class == 'I' and M_chain_seq =='':
            print('No MHC sequence was provided. Trying to retrieve it from reference sequences...')
            try:
                self.retrieve_MHCI_refseq()
            except:
                print('Something went wrong while retrieving the reference sequence.')
                print('Please provide a M_chain_seq for your target.')
                print('###################')
                print('You can find all the reference MHC sequences used in PANDORA')
                print(' and use them for your target in <MyDatabase>.ref_MHCI_sequences')
                print('Where <MyDatabase> is the name of your PANDORA Database object.')
                print('###################')
                print('You can also find reference MHC sequences for Humans at:')
                print('https://www.ebi.ac.uk/ipd/imgt/hla/')
                print('And for other animals here:')
                print('https://www.ebi.ac.uk/ipd/mhc/')
                print('###################')
                raise Exception('Failed in retriving reference sequence.')
        elif MHC_class == 'II' and M_chain_seq =='' and N_chain_seq =='':
            print('No MHC sequence was provided. Trying to retrieve it from reference sequences...')
            raise Exception('Reference MHC II sequences have not been implemented yet. Please provide both M and N chain sequence.')
        
        # If anchors are not provided, predict them from the peptide length
        if MHC_class =='I' and anchors == []:
            #Use Canonical anchors
            if use_netmhcpan == False:
                print('WARNING: no anchor positions provided. Pandora will assign them to canonical anchor position.')
                print('If you want PANDORA to use NetMHCpan to predict the anchors set use_netmhcpan as True')
                anchor_1 = 2
                anchor_2 = len(peptide)
                anchors = [anchor_1, anchor_2]
                self.anchors = anchors
            #Use NetMHCpan to predict the anchors
            else:
                print('WARNING: no anchor positions provided. Pandora will predict them using NetMHCpan')
    
                netMHCpan_dir = [i for i in os.listdir(PANDORA.PANDORA_path + '/../') if
                                 i.startswith('netMHCpan') and os.path.isdir(PANDORA.PANDORA_path + '/../'+i)]
                if os.path.isfile(PANDORA.PANDORA_path + '/../' + netMHCpan_dir[0] + '/netMHCpan'):
                    # predict the anchors
                    self.anchors = Modelling_functions.predict_anchors_netMHCpan(self.peptide, self.allele_type)
                    print('Predicted anchors: %s' %self.anchors)
    
                else:
                    print("Need netMHCIIpan to predict anchor positions. Please download and install netMHCpan.\n\n"
                      "The user needs to manually download the netMHCIIpan software, since it requires agreement to an academic license agreement.\n"
                      "1. Go to: https://services.healthtech.dtu.dk/service.php?NetMHCpan-4.1\n"
                      "2. press the download button\n"
                      "3. Download the most recent version appropriate for your operating system\n"
                      "4. Untar the file and put it in the root directory of your PANDORA install\n"
                      "5. Follow the readme or simply run netMHCpan_install.py to configure netMHCpan")

        if MHC_class =='II' and anchors == []:
            print('WARNING: no anchor positions provided. Pandora will predict them using netMHCIIpan.')

            netMHCIIpan_dir = [i for i in os.listdir(PANDORA.PANDORA_path + '/../') if
                             i.startswith('netMHCIIpan') and os.path.isdir(PANDORA.PANDORA_path + '/../'+i)]
            if os.path.isfile(PANDORA.PANDORA_path + '/../' + netMHCIIpan_dir[0] + '/netMHCIIpan'):
                # predict the anchors
                self.anchors = Modelling_functions.predict_anchors_netMHCIIpan(self.peptide, self.allele_type)
            else:
                print("Need netMHCpan to predict anchor positions. Please download and install netMHCIIpan.\n\n"
                      "The user needs to manually download the netMHCpan software, since it requires agreement to an academic license agreement.\n"
                      "1. Go to: https://services.healthtech.dtu.dk/service.php?NetMHCpan-4.1\n"
                      "2. press the download button\n"
                      "3. Download the most recent version appropriate for your operating system\n"
                      "4. Untar the file and put it in the root directory of your PANDORA install\n"
                      "5. Follow the readme or simply run netMHCpan_install.py to configure netMHCIIpan")


    def info(self):
        """ Print the basic info of this structure

        """
        print('This is a %s structure.' % (type(self).__name__))
        print('ID: %s' % self.id)
        print('Type: MHC class %s' % self.MHC_class)
        print('Alleles: %s' % self.allele_type)
        if self.M_chain_seq != '':
            print('Alpha chain length: %s' %len(self.M_chain_seq))
        if self.N_chain_seq != '' and self.MHC_class == 'II':
            print('Beta chain length: %s' %len(self.N_chain_seq))
        print('Peptide length: %s' %len(self.peptide))
        if self.M_chain_seq != '':
            print('Alpha chain: %s' % self.M_chain_seq)
        if self.N_chain_seq != '':
            print('Beta chain: %s' % self.N_chain_seq)
        print('Peptide: %s' % self.peptide)
        print('Anchors: %s' % self.anchors)
        if self.sheet:
            print('Beta-sheet: %s' % self.sheet)
        if self.helix:
            print('Alpha-helix: %s' % self.helix)
        if self.templates:
            print('Using template %s for homology modelling' %self.templates)
        if self.initial_model:
            print('An initial model has been provided.')

    def calc_contacts(self):
        if self.initial_model:
            self.contacts = Contacts.Contacts(self.initial_model)
        else:
            raise Exception('Provide a PDB structure to the Template object first')

    def calc_anchor_contacts(self):
        if self.initial_model and self.anchors:
            self.anchor_contacts = Contacts.Contacts(self.initial_model, anchors=self.anchors).anchor_contacts
        else:
            raise Exception('Provide an initial model (.ini PDB) and anchor positions to the Target object first')

    def retrieve_MHCI_refseq(self, input_file = None):
        """
        Retrieves MHC I reference sequence from fasta file.

        Args:
            input_file (str, optional): Path to the input reference fasta file. Defaults to None.

        Returns:
            None.

        """
        # Import necessary package
        from Bio import SeqIO
        
        # Define correct fasta file
        if input_file == None:
            if self.allele_type[0].startswith('HLA'):
                input_file = PANDORA.PANDORA_data+ '/csv_pkl_files/Human_MHC_data.fasta'
            else:
                input_file = PANDORA.PANDORA_data+ '/csv_pkl_files/NonHuman_MHC_data.fasta'
        
        # Parse Fasta file
        fasta_sequences = SeqIO.parse(input_file,'fasta')
        
        # Return the right sequences
        seq_flag = False
        for seq in fasta_sequences:
            if seq.id == self.allele_type[0]:
                self.M_chain_seq = str(seq.seq)
                seq_flag = True
                break
            else:
                pass
        
        if seq_flag == True:
            print('MHC sequence correctly retrieved')
        else:
            print('WARNING: No MHC seq could be retrieved with the given MHC allele name')
            print("The model will be generated using the best template's MHC sequence.")
            print("To be sure you use the right sequence, please double check your MHC allele name")
            print("Or provide the MHC sequence as target.M_chain_seq")



#!/usr/bin/env python
import modeller as M
from modeller import automodel as MA       # Load the automodel class
env = M.environ()

class MyLoop(MA.loopmodel):
    def special_patches(self, aln):
        # Rename both chains and renumber the residues in each
        self.rename_segments(segment_ids=['M', 'P'], renumber_residues=[1, 1])
        
    ### Skipping randomness in Initial modelling. Uncomment this funtion only if you specifically mean it. 
    def build_ini_loop(self, atmsel):
        pass
    ###
    
    def select_loop_atoms(self):
        # Two residue ranges (both will be refined simultaneously)
        return M.selection(self.residue_range('%i:P', '%i:P'))# Be sure those arguments are always in the same line!
        # note: the residue numbers and chain IDs refer to the built model, not to the template(s).

    ### SPECIAL_RESTRAINTS_BREAK ###  DO NOT DELETE THIS COMMENT
    
    def special_restraints(self, aln):
        rsr = self.restraints
        atoms = self.atoms
        modelling_stdev= float('%s')  #STDEV MARKER   DO NOT DELETE THIS COMMENT

        # need to add more distance restraints later :)
        contact_file = open('contacts_%s.list', 'r')
        for contact_data_line in contact_file:
            contact_data = (contact_data_line.replace(' ', '')).split('\t')
            rsr.add(M.forms.gaussian(group=M.physical.xy_distance,feature=M.features.distance(atoms['%s:%s:%s' %(contact_data[3], contact_data[2], contact_data[1])], atoms['%s:%s:%s' %(contact_data[7], contact_data[6], contact_data[5])]),mean=float(contact_data[8]), stdev=modelling_stdev))

        contact_file.close()

        # Central Alpha helix
        # ALPHA-HELIX-MARKER  ### DO NOT DELETE THIS COMMENT

        # Central Beta hairpin
        # BETA-SHEET-MARKER  ### DO NOT DELETE THIS COMMENT


from PANDORA.Contacts import Contacts

def pMHCI_anchors(pdb):
    ''' Finds peptide anchor residues of p:MHCI complexes

    Args:
        pdb: Bio.PDB object (or path to pdb file (string))

    Returns: (list of ints) Peptide anchor residue positions

    '''

    def closest_pept_res(dist_list):
        '''Find the closest peptide residue based on te minimal mean value between peptide residues and MHC residues'''
        anch_list = []
        for i in set([i[2] for i in dist_list]):
            l = [x[-1] for x in dist_list if x[2] == i]
            anch_list.append((i, sum(l) / len(l))) #add mean distance between pept residue and MHC residue
        return min(((x[1], x[0]) for x in anch_list), default=(0, 0))[1] #get the pept residue with the min mean dist

    # Rafaella's residues
    # pocket_M = {'anch1': [24, 25, 35, 36], 'anch2': [81]}

    # Derek's modified version. Empirically tested to have the highest prediction accuracy. See below. (only best shown)
    pocket_M = {'anch1': [7, 24, 35], 'anch2': [118, 135]}

    # Accuracies peptide residue anchor prediction using MHCI residue specified above.
    # I selected all structures with non-canonically spaced anchors from Rafaella's prediction and 100 random pdbs.
    # From this dataset I manually determined the peptide anchor residue position. I then looped through all MHCI
    # residues and selected the ones with the highest accuracy in predicting the correct anchors.

    # Getting all anchors right --> Accuracy: 100.0
    # Anchor 1: 100.0
    # Anchor 2: 100.0

    # Calculate the contacts with a cutoff of 18. This cutoff because at lower cutoffs, the orientations of side chains
    # of different structures, result in highly different contacts.
    cont = Contacts.Contacts(pdb, pept_contacts=True, M_only=sum(pocket_M.values(), []), cutoff=25).chain_contacts

    # Make sure the first chain is P, second chain is M or N, not the other way around
    cont = cont + [(i[4], i[5], i[6], i[7], i[0], i[1], i[2], i[3], i[8]) for i in cont if i[1] == 'P' or i[5] == 'P']
    cont = list(dict.fromkeys(cont))

    # Find the peptide residues that reside in the M and N pockets
    anch1, anch2 = [], []
    for i in cont:
        if i[1] == 'P' and i[5] == 'M': # Use residues from the M chain to find anchor 1, 3 and 4
            # Check the Alpha, Beta and Gamma carbon molecules of the peptide residues.
            if i[3] == 'CA' or i[3] == 'CB' or i[3] == 'CG' or i[3] == 'C1':
                if i[6] in pocket_M['anch1']: #If connected (<25 ang) to specified anch 1 MHC residues, add to list
                    anch1.append(i)
                if i[6] in pocket_M['anch2']: #If connected (<25 ang) to specified anch 2 MHC residues, add to list
                    anch2.append(i)

    # Now find the peptide residues that are closest to the specified MHCII residues to get the anchors
    anchor_1 = closest_pept_res(anch1)
    anchor_2 = closest_pept_res(anch2)
    # Put the anchors in the right order if they aren't already
    anchors = sorted([i for i in [anchor_1, anchor_2]])

    # Return the ordered set of anchors. Filter out 0's and >25 numbers if they happen to get through.
    return sorted(set([i for i in anchors if i != 0 or i > 25]))


def pMHCII_anchors(pdb):
    ''' Finds peptide anchor residues of p:MHCII complexes

    Args:
        pdb: Bio.PDB object (or path to pdb file (string))

    Returns: (list of ints) Peptide anchor residue positions

    '''

    def closest_pept_res(dist_list):
        '''Find the closest peptide residue based on te minimal mean value between peptide residues and MHC residues'''
        anch_list = []
        for i in set([i[2] for i in dist_list]):
            l = [x[-1] for x in dist_list if x[2] == i]
            anch_list.append((i, sum(l) / len(l))) #add mean distance between pept residue and MHC residue
        return min(((x[1], x[0]) for x in anch_list), default=(0, 0))[::-1] #get the pept residue with the min mean dist

    # Rafaella's residues
    # pocket_M = {'anch1':[28, 33, 31], 'anch3':[10]}
    # pocket_N = {'anch2':[10, 11, 21, 22, 23], 'anch3':[6,7], 'anch4':[ 28, 27, 31, 33]}

    # Derek's modified version. Empirically tested to have the highest prediction accuracy.
    # Has a 100% accuracy on all IMGT templates
    pocket_M = {'anch1': [28, 29], 'anch2': [82, 81, 80], 'anch4': [82, 81, 80], 'anch3': [113, 114]}
    pocket_N = {'anch2': [24, 25], 'anch3': []}

    cont = Contacts.Contacts(pdb, pept_contacts=True, M_only=sum(pocket_M.values(), []),
                             N_only=sum(pocket_N.values(), []), cutoff=25).chain_contacts

    # Make sure the first chain is P, second chain is M or N, not the other way around
    cont = cont + [(i[4], i[5], i[6], i[7], i[0], i[1], i[2], i[3], i[8]) for i in cont if i[1] == 'P' or i[5] == 'P']
    cont = list(dict.fromkeys(cont))


    pept_len = max([i[2] for i in cont if i[1] == 'P'])

    # Find the peptide residues that reside in the M and N pockets
    anch1, anch2, anch3, anch4 = [], [], [], []
    for i in cont:
        if i[1] == 'P' and i[5] == 'M': # Use residues from the M chain to find anchor 1, 3 and 4
            # Check the Alpha, Beta and Gamma carbon molecules of the peptide residues.
            if i[3] == 'CA' or i[3] == 'CB' or i[3] == 'CG':
                if i[6] in pocket_M['anch1']: #If connected (<25 ang) to specified anch 1 MHC residues, add to list
                    anch1.append(i)
                if i[6] in pocket_M['anch2']: #If connected (<25 ang) to specified anch 2 MHC residues, add to list
                    anch2.append(i)
                if i[6] in pocket_M['anch3']: #If connected (<25 ang) to specified anch 3 MHC residues, add to list
                    anch3.append(i)
                if i[6] in pocket_M['anch4']: #If connected (<25 ang) to specified anch 4 MHC residues, add to list
                    anch4.append(i)

        if i[1] == 'P' and i[5] == 'N': # Use chain N for the second anchor
            # Check the Alpha, Beta and Gamma carbon molecules of the peptide residues
            if i[3] == 'CA' or i[3] == 'CB' or i[3] == 'CG':
                if i[6] in pocket_N['anch2']: #If connected (<25 ang) to specified anch 2 MHC residues, add to list
                    anch2.append(i)

    # First check if the peptide is inverted: C to N term instead of N to C term (this is very rare)
    anchor_1 = closest_pept_res(anch1)
    extra_rules = True
    # If the first anchor is very deep inside the 1st pocket, we can be pretty confident in assigning the other anchors
    if anchor_1[1] < 18:

        # find the distance between res M51 and the predicted first anchor. Make sure its less than 15 angstrom.
        cont_p1 = Contacts.Contacts(pdb, pept_contacts=True, M_only=[51],
                                    N_only=[], cutoff=12).chain_contacts
        # Make sure the first chain is P, second chain is M or N, not the other way around
        cont_p1 = cont_p1 + [(i[4], i[5], i[6], i[7], i[0], i[1], i[2], i[3], i[8]) for i in cont_p1 if
                       i[1] == 'P' or i[5] == 'P']
        cont_p1 = list(dict.fromkeys(cont_p1))
        # Check the Alpha, Beta and Gamma carbon molecules of the peptide residues.
        cont_p1 = [i for i in cont_p1 if i[1] == 'P' and i[5] == 'M' and (i[3] == 'CA' or i[3] == 'CB' or i[3] == 'CG')]

        # Find the distance between residue M51 and the predicted 1st anchor
        dist_between_51_anch1 = closest_pept_res([i for i in cont_p1 if i[2] == anchor_1[0]])[1]

        # If predicted anchor 1 is very close to res M51, then we can be very sure to have found pept res anchor 1
        # deep and firmly inside in the pocket --> assume canonical spacing then.
        if dist_between_51_anch1 != 0 and dist_between_51_anch1 < 12:

            anchor_1 = anchor_1[0]
            anchor_2 = anchor_1 + 3
            anchor_3 = anchor_1 + 5
            if pept_len - anchor_1 >= 8:
                anchor_4 = anchor_1 + 8
            else:
                anchor_4 = 0
            extra_rules = False

    # If its not, use some extra rules to increase prediction accuracy.
    if extra_rules:
        # First find anchor 2, this anchor is found with the highest accuracy
        if pept_len >= 9:
            anch2 = [i for i in anch2 if i[2] < pept_len - 4]
        anchor_2 = closest_pept_res(anch2)[0]

        # Make sure anchor 1 if smaller than anchor 2, but bigger than anchor 2 - 5
        # This makes sure residues from long flanking regions that are flipped over the MHC structure aren't selected
        anch1_mod = [i for i in anch1 if i[2] < anchor_2 - 1 and i[2] > anchor_2 - 5]
        anchor_1 = closest_pept_res(anch1_mod)[0]

        # Make sure anchor 3 is close to anchor 2 as well (but not too far, to prevent selecting anchor 4)
        anch3_mod = [i for i in anch3 if i[2] > anchor_2 + 1 and i[2] < anchor_2 + 4]
        anchor_3 = closest_pept_res(anch3_mod)[0]

        # Same for anchor 4. Make sure its not too far away. Otherwise distant flanking residues might get selected.
        anch4_mod = [i for i in anch4 if i[2] > anchor_2 + 2 and i[2] < anchor_2 + 6]
        anchor_4 = closest_pept_res(anch4_mod)[0]

    # Put the anchors in the right order if they aren't already
    anchors = sorted([i for i in [anchor_1, anchor_2, anchor_3, anchor_4]])

    # If the spacing between 1-4 and 1-2 is both wrong, infer anchors positions based on anchor 1
    if anchors[-1] - anchors[0] != 8 and anchors[1] - anchors[0] != 3:
        anchor_1 = closest_pept_res(anch1)[0]
        if pept_len - anchor_1 >= 8:
            anchors = [anchor_1, anchor_1 + 3, anchor_1 + 5, anchor_1 + 8]
        else:
            anchors = [anchor_1, anchor_1 + 3, anchor_1 + 5, 0]

    # If anchor 1 and 4 are spaced in the canonical way, and anchor 2 and 3 are not, then just jump the gun and
    # use the canonical spacing for anchor 2 and 3.
    if anchors[-1] - anchors[0] == 8:
        if anchors[1] - anchors[2] != -2 or anchors[2] - anchors[1] != 2:
            anchors[1] = anchors[0] + 3
            anchors[2] = anchors[0] + 5

    # If there is the 3rd and 4th anchor are adjacent and the 4th anchor is the last peptide, remove 4th anchor
    if anchors[3] - anchors[2] == 1 and pept_len == anchors[3]:
            anchors[3] = 0

    # Return the ordered set of anchors. Filter out 0's and >25 numbers if they happen to get through.
    return sorted(set([i for i in anchors if i != 0 or i > 25]))



# 23-Aug-2018 14:36

import modeller as M
import modeller.automodel as MA
from MyLoop import MyLoop
import sys

M.log.verbose()                                # request verbose output
env = M.environ()                              # create a new MODELLER environment to build this model in

# directories for input atom files
env.io.atom_files_directory = ['./']

# Read in HETATM records from template PDBs
env.io.hetatm = True

a = MyLoop(env, alnfile= '%s',
              knowns= '%s', sequence = '%s',           #Be sure those two arguments are always in the same line!
              loop_assess_methods = MA.assess.DOPE)
a.make(exit_stage=2)



# 23-Aug-2018 14:36

import modeller as M
import modeller.automodel as MA
import modeller.parallel  as MP
from MyLoop import MyLoop


j = MP.job() 
for i in range(int('%s')):        ## PARALLEL_JOB_LINE_TO_COMPLETE ## DO NOT DELETE THIS COMMENT!
    j.append(MP.local_slave())    ## PARALLEL_JOB_LINE ## DO NOT DELETE THIS COMMENT!

M.log.verbose()                                # request verbose output
env = M.environ()                              # create a new MODELLER environment to build this model in

# directories for input atom files
#env.io.atom_files_directory = ['.', '../atom_files']
env.io.atom_files_directory = './'

# Read in HETATM records from template PDBs
#env.io.hetatm = True
env.io.water = True

a = MyLoop(env, alnfile = '%s',
              knowns = '%s', sequence = '%s',              #Be sure those two arguments are always in the same line!
              loop_assess_methods = MA.assess.DOPE)


a.starting_model= 1                          # index of the first model
a.ending_model  = int('%s')                  # index of the last model

# Optimization
# CG
a.library_schedule = MA.autosched.slow           # Very thorough VTFM optimization
a.max_var_iterations = 300                    # Select length of optimizations
a.max_molpdf = 1e6                            # do not stop unless obj.func. > 1E6

# Loop Modelling

a.loop.starting_model = 1           # First loop model
a.loop.ending_model   = int('%s')          # Last loop model
a.loop.md_level       = MA.refine.slow   # Loop model refinement level

#MD
a.md_level = MA.refine.slow                      # model refinement level

# Repeat the whole cycle 2 times
#a.repeat_optimization = 2

a.use_parallel_job(j)                        ## PARALLEL_JOB_LINE ## DO NOT DELETE THIS COMMENT!
a.make()                                     # do the actual homology modeling



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from PANDORA.PMHC import PMHC
from PANDORA.Pandora import Pandora
from PANDORA.Wrapper.run_model import run_model
import csv
from joblib import Parallel, delayed

class Wrapper():
    def __init__(self):

        """Pandora wrapper object.

        Args:
            None.

        Returns:
            None.

        """
        self.data_file = ''
        self.db = None
        self.targets = {}
        self.jobs = {}


    def __get_targets_from_file(self, data_file, delimiter='\t', header=True,
                               IDs_col=None, peptides_col=0,
                               allele_col=1, anchors_col=None,
                               M_chain_col=None, N_chain_col=None,
                               start_row=None, end_row=None):
        """
        Extracts peptide sequences, alleles and anchors (if specified)
            from the target file.
            Default input should be a .tsv file without any header with
            the following structure: peptides_sequence_col \t alleles_name_col

        Args:
            data_file (str): Path to the input tsv/csv file containing targets
                information.
            delimiter (str, optional): data_file delimiter. Do not use
                semicolons (';') as separators. Defaults to '\t'.
            header (bool, optional): If True, assumes the data_file has a
                header line and skips it. If your file has no header line,
                set it as False. Defaults to True.
            IDs_col (int or None, optional): Column of data_file containing
                the targets IDs. If None, will automatically assign an ID
                according to the row number. Defaults to None.
            peptides_col (int, optional): Column of data_file containing
                the targets peptides. Defaults to 0.
            allele_col (int, optional): Column of data_file containing
                the targets alleles. Umbiguous allele cases (where the allele
                might have multiple names) should be separated by a
                semicolon (';'). Defaults to 1.
            anchors_col (None or int, optional): Column of data_file containing
                the targets anchors. Anchors should be two numbers separated
                by a semicolon (';'). Defaults to 2.
            M_chain_col (None or int, optional): Column of data_file containing
                the targets M chain sequences.
            N_chain_col (None or int, optional): Column of data_file containing
                the targets N chain sequences (only for MHCII).
            start_row (None or int, optional): Starting row of data_file, to use when
                splitting the data_file into multiple batches. This allows to
                specify from which row the samples for this job start.
            end_row (None or int, optional): Ending row of data_file, to use when
                splitting the data_file into multiple batches. This allows to
                specify at which row the samples for this job end.

        Returns:
            None.

        """

        targets = {}
        with open(data_file, 'r') as infile:
            spamreader = csv.reader(infile, delimiter=delimiter)
            if header == True:
                next(spamreader)
            for i, row in enumerate(spamreader):
                if start_row != None and i < start_row:
                    pass
                elif end_row != None and i >= end_row:
                    break
                else:
                    ## Assign target ID
                    if IDs_col != None:
                        target_id = row[IDs_col]
                    else:
                        target_id = 'Target_%i' %(i+1)

                    ## Assign peptide sequence
                    peptide_seq = row[peptides_col]

                    ## Assign allele name
                    allele = row[allele_col].split(';')

                    ## Make target entry
                    targets[target_id] = {'peptide_sequence' : peptide_seq,
                                              'allele' : allele}

                    ## Assign optional arguments. Be sure the empty values correspond
                    ## to the default values in PMHC.Target.__init__()
                    ## Assign anchors
                    if anchors_col:
                        anchors = tuple([int(x) for x in row[anchors_col].split(';')])
                        targets[target_id]['anchors'] = anchors
                    else:
                        targets[target_id]['anchors'] = []

                    ## Assign M chain sequence
                    if M_chain_col:
                        M_chain_seq = row[M_chain_col]
                        targets[target_id]['M_chain_seq'] = M_chain_seq
                    else:
                        targets[target_id]['M_chain_seq'] = ''

                    ## Assign N chain sequence
                    if N_chain_col:
                        N_chain_seq = row[N_chain_col]
                        targets[target_id]['N_chain_seq'] = N_chain_seq
                    else:
                        targets[target_id]['N_chain_seq'] = ''

        self.targets = targets

    def create_targets(self, data_file, database, MHC_class, delimiter = '\t',
                       header=True, IDs_col=None, peptides_col=0, allele_col=1,
                       anchors_col=None, M_chain_col=None, N_chain_col=None,
                       benchmark=False, verbose=False, start_row=None,
                       end_row=None, use_netmhcpan=False):
        """


        Args:
            data_file (str): Path to the input tsv/csv file containing targets
                information.
            database (PANDORA.Database.Database): Database object.
            MHC_class (str): MHC class of the targets, as 'I' or 'II'.
            delimiter (str, optional): data_file delimiter. Do not use
                semicolons (';') as separators. Defaults to '\t'.
            header (bool, optional): If True, assumes the data_file has a
                header line and skips it. If your file has no header line,
                set it as False. Defaults to True.
            IDs_col (int or None, optional): Column of data_file containing
                the targets IDs. If None, will automatically assign an ID
                according to the row number. Defaults to None.
            peptides_col (int, optional): Column of data_file containing
                the targets peptides. Defaults to 0.
            allele_col (int, optional): Column of data_file containing
                the targets alleles. Umbiguous allele cases (where the allele
                might have multiple names) should be separated by a
                semicolon (';'). Defaults to 1.
            anchors_col (int, optional): Column of data_file containing
                the targets anchors. Anchors should be two numbers separated
                by a semicolon (';'). Defaults to 2.
            M_chain_col (None or int, optional): Column of data_file containing
                the targets M chain sequences.
            N_chain_col (None or int, optional): Column of data_file containing
                the targets N chain sequences (only for MHCII).
            benchmark (bool, optional): Set True only for benchmarking purpose,
                if target structures are available. Defaults to False.
            start_row (None or int): Starting row of data_file, to use when
                splitting the data_file into multiple batches. This allows to
                specify from which row the samples for this job start.
            end_row (None or int): Ending row of data_file, to use when
                splitting the data_file into multiple batches. This allows to
                specify at which row the samples for this job end.
            use_netmhcpan (bool, optional): If True, uses local installation
                of netMHCPan to predict anchor positions for each target.

        Returns:
            None.

        """
        self.data_file = data_file
        self.db = database

        ## Extract targets from data_file
        self.__get_targets_from_file(data_file, delimiter=delimiter,
                                     header=header, IDs_col=IDs_col,
                                     peptides_col=peptides_col, allele_col=allele_col,
                                     anchors_col=anchors_col, M_chain_col=M_chain_col,
                                     N_chain_col=N_chain_col, start_row=start_row,
                                     end_row=end_row)

        ## Create target objects
        jobs = {}
        for target_id in self.targets:
            #try:
                if verbose:
                    print('Target ID: ', target_id)
                    print('Target MHC_class: ', MHC_class)
                    print('Target allele: ', self.targets[target_id]['allele'])
                    print('Target peptide: ', self.targets[target_id]['peptide_sequence'])
                    print('Target M chain seq: ', self.targets[target_id]['M_chain_seq'])
                    if N_chain_col:
                        print('Target N chain seq: ', self.targets[target_id]['N_chain_seq'])
                if verbose:
                    print('Target Anchors: ', self.targets[target_id]['anchors'])
                #try:
                tar = PMHC.Target(target_id, allele_type=self.targets[target_id]['allele'],
                                  peptide=self.targets[target_id]['peptide_sequence'] ,
                                  MHC_class=MHC_class, anchors=self.targets[target_id]['anchors'],
                                  M_chain_seq=self.targets[target_id]['M_chain_seq'],
                                  N_chain_seq=self.targets[target_id]['N_chain_seq'],
                                  use_netmhcpan=use_netmhcpan)
                #except Exception as err:
                #    print('Skipping Target %s at Target object generation step for the following reason:' %target_id)
                #    print(("Exception: {0}".format(err)))
                try:
                    mod = Pandora.Pandora(tar, self.db)
                except Exception as err:
                    print('Skipping Target %s at Pandora object generation step for the following reason:' %target_id)
                    print(("Exception: {0}".format(err)))
                try:
                    mod.find_template(benchmark=benchmark)
                    jobs[target_id] = [tar, mod.template]
                except Exception as err:
                    print('Skipping Target %s at template selection step for the following reason:' %target_id)
                    print(("Exception: {0}".format(err)))
            #except Exception as err:
            #    print('An unidentified problem occurred with Target %s. Please check your target info' %target_id)
            #    print(("Exception: {0}".format(err)))
        self.jobs = jobs

    def run_pandora(self, num_cores=1, n_loop_models=20, n_jobs=None,
                    benchmark=False, output_dir=False, pickle_out=False):
        """Runs Pandora in parallel jobs.


        Args:
            num_cores (int, optional): Number of parallel PANDORA jobs.
                Each one will be sent to a different core. Defaults to 1.
            n_loop_models (int, optional): Number of  loop models.
                Defaults to 20.
            n_jobs (int, optional): Number of parallel MODELLER loop jobs.
                Do not increase further than n_loop_models. Defaults to None.
            benchmark (bool, optional): Set True only for benchmarking purpose,
                if target structures are available. Defaults to False.
            output_dir (str, optional): Output directory path.
                Defaults to False.
            pickle_out (bool, optional): If True, outputs a pickle file
                containing every model object. Defaults to False.

        Returns:
            None.

        """

        for job in self.jobs:
            if output_dir:
                self.jobs[job].extend([n_loop_models, n_jobs, benchmark, pickle_out, output_dir])
            else:
                self.jobs[job].extend([n_loop_models, n_jobs, benchmark, pickle_out])
        Parallel(n_jobs = num_cores, verbose = 1)(delayed(run_model)(job) for job in list(self.jobs.values()))




from Bio.PDB import PDBParser
from Bio.PDB import NeighborSearch

class Contacts:
    def __init__(self, PDB, output_file = False, anchors = False, pept_contacts = False, M_only = list(range(1,182)),
                 N_only = list(range(1,92)), cutoff = 5):
        ''' Calculate atom contacts between different chains of a protein or between peptide anchor residues and the
            rest of the MHC structure.

        :param PDB: (String) or (Bio.PDB) Accepts either the path to the pdb file or a Bio.PDB object
        :param output_file: (String) Path to output file. If anchors are provided, only write the anchors output
        :param anchors: (list) list of anchor positions. If this is provided, the atom contacts between the anchor
                        residues of the peptide and the rest of the structure are calculated
        :param cutoff: Cutoff that considers two atoms contacting. Is 5 Angstrom by default
        '''
        self.pdb_path = ''
        self.cutoff = cutoff
        self.anchor_contacts = []
        self.anchors = anchors
        self.output_file = output_file

        # check input type
        if isinstance(PDB, str): #if its a string, it should be the path of the pdb, then load pdb first
            self.pdb_path = PDB
             # Create a parser object, used to read pdb files
            self.PDB = PDBParser(QUIET=True).get_structure('MHC', PDB)
        else: #Else, its already a bio.pdb object
            self.PDB = PDB

        # Actually find the atom contacts between chains
        if not pept_contacts:
            self.chain_contacts = self.find_chain_contacts(self.PDB, self.cutoff)
        if pept_contacts:
            self.chain_contacts = self.find_pept_chain_contacts(self.PDB, M_only=M_only, N_only=N_only, cutoff=self.cutoff)

        # If the user supplied anchors, calculate the anchor contacts
        if anchors: # Calculate peptide anchor residue - structure contacts
            if isinstance(anchors, list):
                # First find which chain is the peptide chain by looking for the shortest chain
                chain_len = [len(i) for i in self.PDB.get_chains()]
                pept_chain = [i.id for i in self.PDB.get_chains()][chain_len.index(min(chain_len))]
                # Only keep the contacts with
                self.anchor_contacts = [i for i in self.chain_contacts if i[1] == pept_chain and i[2] in self.anchors or
                                        i[5] == pept_chain and i[6] in self.anchors]

        # Write output file if the user wants to
        if output_file:
            # Write output file if a filename is provided
            with open(output_file, 'w') as f:
                if anchors: # If anchors are provided, only write the anchor contact file
                    for i in self.anchor_contacts:
                        f.write('\t'.join('%s' %x for x in i) + '\n')
                else: # Else write all contacts. If you want both, this one already contains the anchor contacts
                      # ofcourse.
                    for i in self.chain_contacts:
                        f.write('\t'.join('%s' %x for x in i) + '\n')


    def find_chain_contacts(self, PDB, cutoff=5):
        ''' Calculate atom distances in a pdb object.

        :param pdb: Bio.PDB object
        :param cutoff: (int) Distance cutoff that determines atom contact, standard is 5 Angstrom
        :param all: (Bool) Calculate all atom contacts or just between chains
        :param specific_chain: (String) Only consider contacts between one specific chain and the rest.
        :param anchors: (list) list of anchor positions. If such a list is given, only the contacts between the atoms of
                                the peptide anchor residues and the rest of the structure are calculated.
        :return: list of tuples of: Res 1, Chain 1, Resnr 1, Atom 1, Res 2, Chainm 2, Resnr 2, Atom 2, Distance 1_2
        '''
        # Calculate distances
        atoms = [i for i in PDB.get_atoms()]
        atom_dist = NeighborSearch(atom_list=atoms).search_all(cutoff)
        out = []
        # Append (Res 1,Chain 1,Resnr 1,Atom 1,Res 2,Chain 2,Resnr 2,Atom 2,Distance 1_2) to list for each contact
        for pair in atom_dist:
            out.append((pair[1].get_parent().resname,
                        pair[1].get_parent().get_parent().id,
                        pair[1].get_parent().id[1],
                        pair[1].get_id(),
                        pair[0].get_parent().resname,
                        pair[0].get_parent().get_parent().id,
                        pair[0].get_parent().id[1],
                        pair[0].get_id(),
                        pair[0] - pair[1]))

        # If one would want all atom contacts (also within the same protein chain), the next line can be commented out
        out = [i for i in out if i[1] != i[5]] # filter out contacts between the same chain

        return out

    def find_pept_chain_contacts(self, PDB, M_only = list(range(1,182)), N_only = list(range(1,92)), cutoff=5):
        ''' Calculate atom distances in a pdb object.

        :param pdb: Bio.PDB object
        :param cutoff: (int) Distance cutoff that determines atom contact, standard is 5 Angstrom
        :param all: (Bool) Calculate all atom contacts or just between chains
        :param specific_chain: (String) Only consider contacts between one specific chain and the rest.
        :param anchors: (list) list of anchor positions. If such a list is given, only the contacts between the atoms of
                                the peptide anchor residues and the rest of the structure are calculated.
        :return: list of tuples of: Res 1, Chain 1, Resnr 1, Atom 1, Res 2, Chainm 2, Resnr 2, Atom 2, Distance 1_2
        '''
        # Calculate distances

        # Select the atoms in the peptide and the G-domain. This speeds up anchor calculation with 40/60%
        atoms = []
        atoms = atoms + [i for i in PDB[0]['P'].get_atoms()]
        if 'N' in [i.id for i in PDB.get_chains()]:  # Check if its MHCI
            for i in PDB[0]['M'].get_residues():  # Take the G-domain of the M chain
                if i.id[1] in M_only:
                    atoms = atoms + [i for i in i.get_atoms()]
            for i in PDB[0]['N'].get_residues():  # Take the G-domain of the N chain
                if i.id[1] in N_only:
                    atoms = atoms + [i for i in i.get_atoms()]
        elif 'N' not in [i.id for i in PDB.get_chains()]:
            for i in PDB[0]['M'].get_residues():  # Take the G-domain of the M chain
                if i.id[1] in M_only:
                    atoms = atoms + [i for i in i.get_atoms()]

        atom_dist = NeighborSearch(atom_list=atoms).search_all(cutoff)
        out = []
        # Append (Res 1,Chain 1,Resnr 1,Atom 1,Res 2,Chain 2,Resnr 2,Atom 2,Distance 1_2) to list for each contact
        for pair in atom_dist:
            out.append((pair[1].get_parent().resname,
                        pair[1].get_parent().get_parent().id,
                        pair[1].get_parent().id[1],
                        pair[1].get_id(),
                        pair[0].get_parent().resname,
                        pair[0].get_parent().get_parent().id,
                        pair[0].get_parent().id[1],
                        pair[0].get_id(),
                        pair[0] - pair[1]))

        # If one would want all atom contacts (also within the same protein chain), the next line can be commented out
        out = [i for i in out if i[1] != i[5]] # filter out contacts between the same chain

        return out

    def show(self):
        """Print the contacts. If anchor contacts have been calculated, only show them"""
        if self.anchor_contacts:
            for i in self.anchor_contacts:
                print(i)
        else:
            for i in self.chain_contacts:
                print(i)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from PANDORA.Pandora import Pandora

def run_model(args):
    """Runs one modelling job. Meant to be runned from Pandora.Wrapper

    Args:
        args (list): List of arguments. Should be containing the following, in
            order.
        target (Pandora.PMHC.PMHC.Target): Target object.
        template (Pandora.PMHC.PMHC.Template): Template object.
        n_loop_models (int, optional): Number of loop models. Defaults to 20.
        benchmark (bool, optional): Set True if running a benchmark to retrieve
            models RMSD with reference structures. Defaults to False.

    Returns:
        None.

    """
    target = args[0]
    template = args[1]
    n_loop_models = args[2]
    n_jobs=args[3]
    benchmark = args[4]
    pickle_out = args[5]

    # Check if the output directory is provided
    if len(args) == 6:
        output_dir = False
    elif len(args) == 7:
        output_dir = args[6]

    # Create Pandora Object
    if output_dir:
        mod = Pandora.Pandora(target, template=template,
                              output_dir=output_dir)
    else:
        mod = Pandora.Pandora(target, template = template)

    # Run the modelling
    mod.model(n_loop_models=n_loop_models, n_jobs=n_jobs,
              stdev=0.1, benchmark=benchmark, pickle_out=pickle_out)



# -*- coding: utf-8 -*-
import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

# To update the package version number, edit PANDORA/__version__.py
version = {}
with open(os.path.join(here, 'PANDORA', '__version__.py')) as f:
    exec(f.read(), version)

with open('README.md') as readme_file:
    readme = readme_file.read()

PACKAGES=[
        "PANDORA"
]

setup(
    name='csb_pandora',
    version=version['__version__'],
    description='Model peptide-MHC I complexes using anchor distance restrains in MODELLER',
    long_description=readme + '\n\n',
    long_description_content_type='text/markdown',
    author='Dario F. Marzella, Farzaneh M. Parizi, Derek van Tilborg, Li Xue',
    url='https://github.com/X-lab-3D/PANDORA/tree/master',
    project_urls={
        'Source Code': 'https://github.com/X-lab-3D/PANDORA/tree/master',
        'Issue tracker': 'https://github.com/X-lab-3D/PANDORA/issues'
    },
    packages=PACKAGES,
    #packages=find_packages(),
    include_package_data=True,
    license="Apache Software License 2.0",
    keywords='PANDORA',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],


    install_requires=[
       'Biopython',
       'pdb_tools',
       'pdb2sql',
       'matplotlib',
       'dill'
       ],

    extras_require={
        'doc': ['recommonmark', 'sphinx', 'sphinx_rtd_theme'],
        'test': ['pytest', 'pytest-runner',
                 'coverage', 'coveralls', 'pycodestyle']
    }
)




import PANDORA
from PANDORA.Pandora import Align
from PANDORA.Pandora import Modelling_functions
import time
import os
from Bio.PDB import PDBParser
from datetime import datetime


class Pandora:

    def __init__(self, target, database=None, template=None, output_dir=PANDORA.PANDORA_data + '/outputs'):
        '''__init__(self, target, database=None, template=None, output_dir=PANDORA.PANDORA_data + '/outputs')
        '''
        self.target = target
        self.template = template
        self.database = database
        self.output_dir = output_dir

        if database is None and template is None:
            raise Exception('Provide a Database object so Pandora can find the best suitable template structure for '
                            'modelling. Alternatively, you can specify a user defined Template object.')

    def find_template(self, best_n_templates=1, benchmark=False, verbose=True):
        ''' Find the best template structure given a Target object

        Args:
            best_n_templates (int, optional): how many template structures are
                used for modelling. The best n are used. Defaults to 1.
            benchmark (bool): Perform L-RMSD calculations? only works if the
                target id is an existing pdb id. Defaults to False.
            verbose: (bool): Print information. Defaults to True.

        '''

        if verbose:
            print('\tTarget MHC Class: %s' % self.target.MHC_class)
            print('\tTarget Allele:  %s' % self.target.allele_type)
            print('\tTarget Peptide: %s' % self.target.peptide)
            print('\tTarget Anchors: %s,%s\n' % (self.target.anchors[0],self.target.anchors[1]))

        if self.template is None: # Only find the best template if the user didn't specify one
            # if verbose and self.target.M_chain_seq != '' and seq_based_templ_selection:
            #     print('\tUsing sequence based template selection')
            if verbose:
                print('\tUsing allele type based template selection')
            # Find the best template. If the target already exists in the database,
            # also consider the initial loop model as a model
            self.template, self.pept_ali_scores, self.keep_IL = Modelling_functions.find_template(self.target,
                                                                    self.database,
                                                                    best_n_templates=best_n_templates,
                                                                    benchmark=benchmark)
            self.target.templates = [i.id for i in self.template]
            if verbose:
                print('\tSelected template structure (%s): %s' %(len(self.template), [i.id for i in self.template]))

        else:
            if verbose:
                if type(self.template)==list:
                    print('\tUser defined template structure (%s): %s' %(len(self.template), [i.id for i in self.template]))
                else:
                    print('\tUser defined template structure: %s' %self.template.id)
            # Check if the target structure and template structure are the same.
            if type(self.template)==list:
                self.keep_IL = any(Modelling_functions.check_target_template(self.target, tmpl) for tmpl in self.template)
            else:
                self.keep_IL = Modelling_functions.check_target_template(self.target, self.template)
            # determine peptide alignment scores of the target and the template(s)
            self.pept_ali_scores = []

            if type(self.template)==list:
                for templ in self.template:
                    if self.target.id == 'I':
                        score = Modelling_functions.score_peptide_alignment_MHCI(self.target, templ, 'PAM30')
                        self.pept_ali_scores.append((score, templ.peptide, templ.id))
                    elif self.target.id == 'II':
                        score = Modelling_functions.score_peptide_alignment_MHCII(self.target, templ, 'PAM30')
                        self.pept_ali_scores.append((score, templ.peptide, templ.id))
            else:
                if self.target.id == 'I':
                    score = Modelling_functions.score_peptide_alignment_MHCI(self.target, self.template, 'PAM30')
                    self.pept_ali_scores.append((score, self.template.peptide, self.template.id))
                elif self.target.id == 'II':
                    score = Modelling_functions.score_peptide_alignment_MHCII(self.target, self.template, 'PAM30')
                    self.pept_ali_scores.append((score, self.template.peptide, self.template.id))
            self.pept_ali_scores = self.pept_ali_scores[:best_n_templates]



        if verbose:
            if type(self.template)==list:
                print('\tTemplates Allele:  %s' %([i.allele_type for i in self.template]))
                print('\tTemplates Peptide: %s' %([i.peptide for i in self.template]))
                print('\tTemplates Anchors: %s\n' %([i.anchors for i in self.template]))
            else:
                print('\tTemplate Allele:  %s' %self.template.allele_type)
                print('\tTemplate Peptide: %s' %self.template.peptide)
                print('\tTemplate Anchors: %s\n' %self.template.anchors)

        #TODO: remove this line only after implementing issue #32.
        if type(self.template)==list:
            self.template = self.template[0]

    def prep_output_dir(self):
        ''' Create an output directory and move the template pdb there
            Uses self.output_dir (str): Path to output directory.
                Defaults to <PANDORA_location>/PANDORA_files/data/outputs.

        Args:
            None


        '''

        # create an output directory
        try:
            self.output_dir = '%s/%s_%s' %(self.output_dir, self.target.id, self.template.id)
            #self.output_dir = '%s/%s_%s' % (self.output_dir, self.target.id, '_'.join([i.id for i in self.template]))

            if not os.path.exists(self.output_dir):
                os.makedirs(self.output_dir)
        except:
            raise Exception('A problem occurred while creating output directory')

        #for templ in self.template:

        if os.path.isfile(self.template.pdb_path):
            os.system('cp %s %s/%s.pdb' %(self.template.pdb_path, self.output_dir, self.template.id))
        else:
            print('Template object could not be found. Please check the path: %s.' %self.template.pdb_path)
            print('If the path is not available, you can use Database.repath.')
            raise Exception('Template file not found.')

        # dd/mm/YY H:M:S
        #date_time = datetime.now().strftime("%d-%m-%Y-%H-%M-%S")
        # Create the output dir of the specific case
        #self.output_dir = '%s/%s_%s_%s' %(output_dir, self.target.id, '_'.join([i.id for i in self.template]), date_time)
        #if not os.path.exists(self.output_dir):
        #    os.makedirs(self.output_dir)

        # copy the template structure to the output file
        #for t in self.template:
        os.system('cp %s %s/%s.pdb' %(self.template.pdb_path, self.output_dir, self.template.id))

    def align(self, verbose=True):
        ''' Create the alignment file for modeller.

        Args:
            verbose: (bool): Print information

        '''
        self.alignment = Align.Align(self.target, self.template, output_dir=self.output_dir)

        # self.alignment = Align.Align2(target = self.target, template=self.template, output_dir=self.output_dir)
        # self.alignment.align_templates()

        if verbose:
            print('\tSuccessfully created alignment file')

    def write_ini_script(self):
        ''' Write the python scipt that modeller uses for creating the initial model'''
        os.chdir(os.path.dirname(PANDORA.PANDORA_path))
        Modelling_functions.write_ini_script(self.target, self.template, self.alignment.alignment_file, self.output_dir)

    def create_initial_model(self, python_script = 'cmd_modeller_ini.py', verbose = True):
        ''' Run modeller to create the initial model. Modeller can only output files in its work directory
            (why though?), so the current work directory is changed to the output dir and later changed back the the
            old working dir.


        Args:
            python_script:  (str): path to script that performs the modeller modelling. Default = cmd_modeller_ini.py
            verbose:  (bool): Print information. Default = True

        '''
        # Change working directory
        os.chdir(self.output_dir)
        # Run Modeller
        os.popen('python %s > modeller_ini.log' %python_script).read()

        try:
            # Load initial model into target object
            self.target.initial_model = PDBParser(QUIET=True).get_structure(self.target.id, self.target.id + '.ini')
        except FileExistsError:
            # If the file does not exist, raise an exception to prompt the user to check MODELLER installation
            raise Exception('.ini file could not be modelled. Please check modeller_ini.log. Is your MODELLER correctly installed?')

        # Change working directory back
        os.chdir(os.path.dirname(PANDORA.PANDORA_path))
        if verbose:
            print('\tSuccessfully created the initital model')

    def run_modeller(self, python_script='cmd_modeller.py', benchmark=False, pickle_out=True, verbose=True,
                     keep_IL=False, RMSD_atoms=['C', 'CA', 'N', 'O']):
        ''' Perform the homology modelling of a target structure on template model(s). Models are saved in the output
            directory and in pandora.results[].

        Args:
            python_script: (str): path to script that performs the modeller modelling. Default = cmd_modeller.py
            benchmark: (bool): Perform L-RMSD calculations? only works if the target id is an existing pdb id.
                                Default = False
            pickle_out: (bool): Save a .pkl with the results. Default = True
            verbose: (bool): Print information. Default = True
            keep_IL: (bool): Keep the initial homology model (non optimized loops). Default = False
            RMSD_atoms: (list[str]): atoms used for the L-RMSD calculation. Default = ['C', 'CA', 'N', 'O'], which is
                                    the backbone

        '''

        if verbose:
            if type(self.template)==list:
                print('\tPerforming homology modelling of %s on %s...' %(self.target.id, '_'.join([t.id for t in self.template])))
            else:
                print('\tPerforming homology modelling of %s on %s...' %(self.target.id, self.template.id))
        t0 = time.time()
        self.results = Modelling_functions.run_modeller(self.output_dir, self.target, python_script=python_script,
                                                        benchmark=benchmark, pickle_out=pickle_out, keep_IL=keep_IL,
                                                        RMSD_atoms=RMSD_atoms)
        if verbose:
            print('\n\tModelling was successfull and took %s seconds' %(round(time.time() - t0, 2)))

    def anchor_contacts(self, verbose=True):
        ''' Calculate anchor contacts and writes a contacts.list file that modeller uses for restraints.

        Args:
            verbose: (bool): Print information. Default = True

        '''

        if verbose:
            print('\tCalculating peptide anchor residue constraints...')
        self.target.calc_anchor_contacts()
        #    Write output file
        with open(self.output_dir + '/contacts_' + self.target.id + '.list', 'w') as f:
                for i in self.target.anchor_contacts:
                    f.write('\t'.join('%s' % x for x in i) + '\n')

    def write_modeller_script(self, n_loop_models=20, n_homology_models = 1, loop_refinement='slow',
                              n_jobs=None, stdev=0.1, helix=False, sheet=False):
        ''' Write the script that modeller uses for the final homology modelling. Most modelling settings are set in
            this script.

        Args:
            n_loop_models: (int): number of loop refinement models PANDORA will generate. Default = 20
            n_homology_models: (int): number of generated homology models PANDORA generates. Default = 1
            loop_refinement: (str): levels of loop refinements. Default = slow. Supported: very_fast, fast, slow,
                                    very_slow, slow_large.
            n_jobs: (int): number of parallel jobs. Is recommended to use as many jobs as the number of models:
            less will result in a slower run, more will not add any benefit but might occupy cores unnecessarily.
            stdev: (float): standard deviation of modelling restraints that is allowed during modelling. Default = 0.1.
                A stdev of 0.2 is recommended for modelling MHC II
            helix: (bool/list): False if no alpha-helix must be modelled. Otherwise, a list of the alpha helix start
                and end-positions as integers. I.e. [3,8] for a helix between peptide residue 3 and 8.
            sheet: (bool/list): False if no beta-sheet must be modelled. Otherwise, a list containing: start position
                of B-sheet 1, start position of B-sheet 2 and the length of the B-sheet in h-bonds. For example:
                ["O:2:P","N:54:M",2] for a parallel B-sheet; The sheet starts at the Oxigen atom of the 2nd residue of
                chain P and at the Nitrogen of the 54th residue of chain M and has a length of 2 H-bonds. Or;
                ["N:6:P", "O:13:P", -3], with -3 denoting an anti-parallel B-sheet with a length of 3 H-bonds.

        '''

        Modelling_functions.write_modeller_script(self.target, self.template, self.alignment.alignment_file,
                                                  self.output_dir, n_loop_models=n_loop_models,
                                                  n_homology_models=n_homology_models, loop_refinement=loop_refinement,
                                                  n_jobs=n_jobs, stdev=stdev, helix=helix, sheet=sheet)

    def __log(self, target_id, template_id, error, logfile = PANDORA.PANDORA_data + '/outputs/Pandora_log.txt', verbose=True):
        ''' Keeps track of what goes wrong while parsing

        Args:
            target_id: (str): ID of target structure
            template_id: (str): ID of template structure
            error: (str): error to append to log file
            logfile: (str): path to logfile
            verbose: (bool): print error?
        '''

        # Create log file
        if not os.path.exists(logfile):
            with open(logfile, 'w') as f:
                f.write('Target\tTemplate\tError\n')

        if verbose:
            print('\t' + error)
        with open(logfile, 'a') as f:
            f.write('%s\t%s\t%s\n' % (target_id, template_id, error))

    def model(self, n_loop_models=20, n_homology_models=1,
              best_n_templates=1, n_jobs=None, loop_refinement='slow', pickle_out=False,
              stdev=0.1, benchmark=False, verbose=True, helix=False, sheet=False, RMSD_atoms=['C', 'CA', 'N', 'O']):
        '''model(self, output_dir=PANDORA.PANDORA_data + '/outputs', n_loop_models=20, n_homology_models=1, best_n_templates=1, n_jobs=None, loop_refinement='slow', pickle_out=False,stdev=0.1, benchmark=False, verbose=True, helix=False, sheet=False, RMSD_atoms=['C', 'CA', 'N', 'O'])
        Wrapper function that combines all modelling steps.

        Args:
            benchmark: (Optional, bool) If True, performs L-RMSD calculations with target strcutre.
                Only works if the target id is present in the template set.
                Defaults to False.
            helix (Optional, False or list): List of integers. Contains starting and ending
                position of a predicted alpha-helix in the peptide. Defaults to False.
            loop_refinement (Optional, str): Type of MODELLER loop refinement to apply.
                Available options are: very_fast,fast,slow,very_slow,slow_large.
                Defaults to 'slow'.
            n_loop_models (Optional, int): number of models modeller generates per run.
                Defaults to 20.
            n_homology_models (Optional, int): number of initial peptide homology models to generate.
                Defaults to 1.
            n_jobs (Optional, int or None): Number of parallel loop model jobs.
                Setting it higher than n_loop_models gives no computational time advantage.
                Recommended to change only when producing high number of loop models
                for one peptide. Defaults to None.
            output_dir (Optional, str): Path to output directory.
                Defaults to PANDORA.PANDORA_data + '/outputs'.
            pickle_out (Optional, bool): If True, saves a pickle file containing the
                PANDORA.PMHC.Model objects for the generated models in the
                output directory. Defaults to False.
            RMSD_atoms (Optional, list): list of atoms to use for final RMSD calculation.
                Works only if benchmark==True. Defaults to ['C', 'CA', 'N', 'O']
            stdev (Optional, float): standard deviation of modelling restraints.
                A higher stdev means more flexible restraints. Defaults to 0.1.
            sheet (Optional, False or list): List containing: start position of B-sheet 1,
                start position of B-sheet 2 and the length of the B-sheet in h-bonds.
                For example: ["O:2:P","N:54:M",2] for a parallel B-sheet; The sheet starts
                at the Oxigen atom of the 2nd residue of chain P and at the Nitrogen of the 54th residue of
                chain M and has a length of 2 H-bonds. Or; ["N:6:P", "O:13:P", -3], with -3 denoting an
                anti-parallel B-sheet with a length of 3 H-bonds.
            verbose (Optional, bool): If True, print modelling information. Defaults to True.

        Returns:
            None

        '''

        if verbose:
            print('\nModelling %s...\n' %self.target.id)

        # Make sure we're in the root directory
        os.path.dirname(PANDORA.PANDORA_path)

        # Find the best template structure given the Target
        try:
            self.find_template(best_n_templates=best_n_templates, benchmark=benchmark, verbose=verbose)
        except:
            self.__log(self.target.id, 'None', 'Could not find a template')
            raise Exception('Could not find a template')

        print('###############')
        print('TEMPLATE: ', self.template.id)
        # Prepare the output directory
        try:
            self.prep_output_dir()
        except:
            self.__log(self.target.id, self.template.id, 'Failed creating output directory')
            raise Exception('Failed creating output directory')



        # Perform sequence alignment. This is used to superimpose the target on the template structure in later steps
        try:
            self.align(verbose=verbose)
        except:
            self.__log(self.target.id, self.template.id, 'Failed aligning target and template')
            raise Exception('Failed aligning target and template')

        # Prepare the scripts that run modeller
        try:
            self.write_ini_script()
        except:
            self.__log(self.target.id, self.template.id, 'Failed writing .ini script')
            raise Exception('Failed writing .ini script')

        # Run modeller to create the initial model
        try:
            self.create_initial_model(verbose=verbose)
        except Exception:
            self.__log(self.target.id, self.template.id, 'Failed creating initial model with modeller')
            raise Exception('Failed creating initial model with modeller')

        # Calculate anchor restraints
        try:
            self.anchor_contacts(verbose=verbose)
        except:
            self.__log(self.target.id, self.template.id, 'Failed calculating anchor restraints')
            raise Exception('Failed calculating anchor restraints')

        # prepare the scripts that run modeller
        try:
            self.write_modeller_script(n_loop_models=n_loop_models, n_homology_models=n_homology_models,
                                       loop_refinement=loop_refinement, n_jobs=n_jobs,
                                       stdev=stdev, helix=helix, sheet=sheet)
        except:
            self.__log(self.target.id, self.template.id, 'Failed preparing the modeller script')
            raise Exception('Failed preparing the modeller script')

        # Do the homology modelling
        try:
            self.run_modeller(benchmark=benchmark, verbose=verbose, keep_IL=self.keep_IL,
                              RMSD_atoms=RMSD_atoms, pickle_out=pickle_out)
        except:
            self.__log(self.target.id, self.template.id, 'Failed running modeller')
            raise Exception('Failed running modeller')


        # if verbose and benchmark:
        #     try:
        #         print('\n\tModel\t\t\t\tMolpdf\t\tL-RMSD\t\tcore L-RMSD')
        #         molsort = sorted(self.results, key=lambda m: float(m.molpdf))
        #         for m in molsort:
        #             try:
        #                 print('\t%s\t\t%s\t\t%s\t\t%s' % (
        #                     os.path.basename(m.model_path).replace('.pdb', ''), round(float(m.molpdf), 4),
        #                     round(float(m.lrmsd), 4), round(float(m.core_lrmsd), 4)))
        #             except AttributeError:
        #                 try:
        #                     print('\t%s\t\t%s\t\t%s' % (
        #                         os.path.basename(m.model_path).replace('.pdb', ''), round(float(m.molpdf), 4),
        #                         round(float(m.lrmsd), 4)))
        #                 except AttributeError:
        #                     print('\t%s\t\t%s' % (
        #                         os.path.basename(m.model_path).replace('.pdb', ''), round(float(m.moldpf), 4)))


        #     except:
        #         self.__log(self.target.id, self.template.id, 'Could not calculate L-RMSD')
        #         raise Exception('Could not calculate L-RMSD')

        # elif verbose and not benchmark:
        if verbose:
            print('\n\tModel\t\t\t\tMolpdf')
            for m in self.results:
                print('\t%s\t\t%s' %(os.path.basename(m.model_path).replace('.pdb', ''), round(float(m.molpdf), 4)))

        if type(self.template)==list:
            self.__log(self.target.id, self.template.id, 'Successfully modelled %s models' %(n_homology_models*n_loop_models))
        else:
            self.__log(self.target.id, self.template.id, 'Successfully modelled %s models' %(n_homology_models*n_loop_models))



import PANDORA
import pickle
from PANDORA.PMHC import PMHC
from PANDORA.Database import Database_functions
import os

class Database:

    def __init__(self):
        self.MHCI_data = {}
        self.MHCII_data = {}
        self.ref_MHCI_sequences = {}
        self.__IDs_list_MHCI = []
        self.__IDs_list_MHCII = []

    def download_data(self, data_dir = PANDORA.PANDORA_data, download = True):
        """download_data(self, data_dir = PANDORA.PANDORA_data, download = True)
        Download all MHC structures and get a two lists that contains all MHCI and MHCII IDs respectively"""

        if download:
            print('Downloading structures ...')
            Database_functions.download_unzip_imgt_structures(data_dir = data_dir, del_inn_files=True, del_kabat_files=True)
        self.__IDs_list_MHCI = Database_functions.download_ids_imgt('MH1', data_dir = data_dir, out_tsv='all_MHI_IDs.tsv')
        self.__IDs_list_MHCII = Database_functions.download_ids_imgt('MH2', data_dir = data_dir, out_tsv='all_MHII_IDs.tsv')


    def __clean_MHCI_file(self, pdb_id, data_dir):
        """ Clean all MHCI structures"""
        return Database_functions.parse_pMHCI_pdb(pdb_id,
                                                   indir = data_dir + '/PDBs/IMGT_retrieved/IMGT3DFlatFiles',
                                                   outdir = data_dir + '/PDBs/pMHCI',
                                                   bad_dir = data_dir + '/PDBs/Bad/pMHCI')

    def __clean_MHCII_file(self, pdb_id, data_dir):
        """ Clean all MHCII structures. Returns a list of bad PDBs"""
        return Database_functions.parse_pMHCII_pdb(pdb_id,
                                                   indir = data_dir + '/PDBs/IMGT_retrieved/IMGT3DFlatFiles',
                                                   outdir = data_dir + '/PDBs/pMHCII',
                                                   bad_dir = data_dir + '/PDBs/Bad/pMHCII')

    def update_ref_sequences(self):
        """Downloads and parse HLA and other MHC sequences to compile reference fastas.
        Returns a dictionary that can be used to select the desired reference sequence"""
        self.ref_MHCI_sequences = Database_functions.generate_mhcseq_database()

    def construct_database(self, save, data_dir = PANDORA.PANDORA_data,
                           MHCI=True, MHCII=True, download=True,
                           update_ref_sequences=True):
        '''construct_database(self, save, data_dir = PANDORA.PANDORA_data, MHCI=True, MHCII=True, download=True, update_ref_sequences=True)
        Construct the database. Download, clean and add all structures

        Args:
            save (str/bool): Filename of database, can be False if you don't want to save the database
            data_dir (str): Path of data directory. Defaults to PANDORA.PANDORA_data.
            MHCI (bool): Parse data for MHCI. Defaults to True.
            MHCII (bool): Parse data for MHCII. Defaults to True.
            download (bool): If True, download the structures data from IMGT. Defaults to True.
            update_ref_sequences (bool): If True, downloads and parse reference sequence strcutres. Defaults to True

        Returns: Database object

        '''
        # Download the data
        self.download_data(download = download, data_dir = data_dir)

        # Construct the MHCI database
        if MHCI:
            # Parse all MHCI files
            for id in self.__IDs_list_MHCI:
                try:
                    templ = self.__clean_MHCI_file(pdb_id = id, data_dir = data_dir)
                    if templ != None:
                        self.MHCI_data[id] = templ
                except:
                    pass

        # Construct the MHCII database
        if MHCII:
            # Parse all MHCII files
            for id in self.__IDs_list_MHCII:
                try:
                    templ = self.__clean_MHCII_file(pdb_id = id, data_dir = data_dir)
                    if templ != None:
                        self.MHCII_data[id] = templ
                except:
                    print('something went wrong parsing %s' %id)

        if save:
            self.save(save)

        #Download and parse HLA and MHC sequences reference data
        if update_ref_sequences:
            self.update_ref_sequences()

        print('Database correctly generated')

    def add_structure(self, id, allele_type, peptide = '', MHC_class = 'I', chain_seq = [], anchors = [], pdb_path = False, pdb = False):
        ''' Add a single structure to the database

        Args:
            id: (str) PDB identifier
            allele_type: (lst) list of MHC alleles (or allele)
            peptide: (str) peptide sequence
            MHC_class: (str) either 'I' or 'II' denoting MHC class I and MHC class II respectively
            chain_seq: (lst) list of chain sequence(s) for the M and N (Alpha and Beta) chain respectively
            anchors: (lst) list of integers specifying which residue(s) of the peptide should be fixed as an anchor
                        during the modelling. MHC class I typically has 2 anchors, while MHC class II typically has 4.
            pdb_path: (str) path to pdb file
            pdb: (Bio.PDB) Biopython PBD object

        '''

        if not id:
            if not pdb_path or not pdb:
                raise ValueError('Structure id or path of .pdb files was not given. Enter value for id and pdb_path')
        # Add to MHCI data
        if MHC_class == 'I':
            self.MHCI_data[id] = PMHC.Template(id, allele_type, peptide, MHC_class, chain_seq, anchors, pdb_path, pdb)
        # Add to MHCII data
        if MHC_class == 'II':
            self.MHCII_data[id] = PMHC.Template(id, allele_type, peptide, MHC_class, chain_seq, anchors, pdb_path, pdb)

    def remove_structure(self, id =''):
        ''' Removes a structure (by id) from the database

        Args:
            id: (str) PDB ID

        '''

        # Remove structure from database
        self.MHCI_data.pop(id, None)
        self.MHCII_data.pop(id, None)

    def repath(self, new_folder_path, save):
        """
        Necessary if the absolut path to the templates structures is different
        from the one used while generating the database.
        It changes the template.pdb_path for each template object in the database
        and returns the modified database.

        Args:
            new_folder_path (str): New path to the 'PDBs' directory contaning template structures.
            save (str/bool): If False, doesn't save the modified database. If str, saves the modified database to the specified file path.'

        Returns:
            None.

        Example:
            >>> MyDatabase.repath('/home/Users/MyUserName/PANDORA/PDBs/', './MyHome_Database.pkl')

        """

        if type(new_folder_path) != str:
            raise Exception('Non-string argument detected. Please provide a valid path as argument.')

        if self.MHCI_data != {}:
            for id in self.MHCI_data:
                from_pMHCI_path = os.path.join(*os.path.normpath(self.MHCI_data[id].pdb_path).split('/')[-2:])
                self.MHCI_data[id].pdb_path = os.path.join(new_folder_path, from_pMHCI_path)

        if self.MHCII_data != {}:
            for id in self.MHCII_data:
                from_pMHCII_path = os.path.join(*os.path.normpath(self.MHCII_data[id].pdb_path).split('/')[-2:])
                self.MHCII_data[id].pdb_path = os.path.join(new_folder_path, from_pMHCII_path)

        if save:
            self.save(save)

    def save(self, fn = 'db.pkl'):
        """Save the database as a pickle file

        :param fn: (str) pathname of file
        """
        with open(fn, "wb") as pkl_file:
            pickle.dump(self, pkl_file)

def load(file_name):
    """Loads a pre-generated database


    Args:
        file_name (str): Dabase file name/path.

    Returns:
        Database.Database: Database object.

    Example:
        >>> db = Database.load('MyDatabase.pkl')

    """
    with open(file_name, 'rb') as inpkl:
        db = pickle.load(inpkl)
    return db
